"""Mambu Entities

.. autosummary::
   :nosignatures:
   :toctree: _autosummary
"""

import copy
from importlib import import_module
import json
import time

from .classes import GenericClass
from .interfaces import (
    MambuWritable,
    MambuAttachable,
    MambuSearchable,
    MambuCommentable,
    MambuOwnable,
)
from .connector.rest import MambuConnectorREST
from .mambustruct import MambuStruct
from .vos import MambuDocument, MambuComment, MambuValueObject
from MambuPy.mambuutil import MambuError, MambuPyError, setup_logging


logger = setup_logging(__name__)


class MambuEntity(MambuStruct):
    """A Mambu object that you may work with directly on Mambu web too."""

    _prefix = ""
    """prefix constant for connections to Mambu"""

    _connector = None
    """Default connector (REST)"""

    _filter_keys = []
    """allowed filters for get_all filtering"""

    _sortBy_fields = []
    """allowed fields for get_all sorting"""

    def __init__(self, **kwargs):
        super().__init__(cf_class=MambuEntityCF, **kwargs)
        if "connector" not in kwargs:
            self._connector = MambuConnectorREST(**kwargs)
        else:
            self._connector = kwargs["connector"]

    def __search_field_in_cfsets(self, field):
        """Search for a field in custom field sets for the entity.

        Retrieves all the customfieldsets for this entity and looks if the
        field lives there.

        If field is found, returns the path going through the set.

        Args:
          field (str): the field to search in customfieldsets

        Returns:
          (str): the path for the field when found or None if not found.
        """
        if not hasattr(self, "_mcfs"):
            from .mambucustomfield import MambuCustomFieldSet

            owner_type = getattr(self, "_ownerType", "")
            self._mcfs = MambuCustomFieldSet.get_all(availableFor=owner_type)
        for cfs in self._mcfs:
            for cf in cfs.customFields:
                if field == cf["id"]:
                    return "/" + cfs.id + "/" + field

    def _extract_field_path(self, field, attrs_dict={}, original_keys=[], cf_class=None):
        """Extracts the path for a given field.

        If the field is a Custom Field, the path is given in a property in its
        attrs dict. (cf `MambuPy.api.entities.MambuEntityCF`)

        If not, first it looks at the original keys in attrs to look if it
        lives in the root (/) of the entity.

        If not, looks for the field in the custom field sets of the entity.

        If not, the path does not exists, but it may be a valid path,
        default to / . Here we are hoping the path is correct,
        otherwise a PATCH may throw an error.

        TODO: perhaps here, in the default case, a better approach
        would be to implement, and maintain, the schema for each
        entity as a VO, and validate if field belongs here. If not,
        default would be an empty string for not being a valid field.

        Args:
          field (str): a field for which to extract the path
          attrs (dict): an attrs structure holding the field
          original_keys (list): keys for entity original attrs
          cf_class (obj): the class used for custom field VOs. If the field is
                          a CF in attrs, the path is a property of it. If not,
                          it lives on the "root" of the attrs

        Returns:
          (str): the path to the field.

        """
        if not cf_class:
            cf_class = self._cf_class
        if not attrs_dict:
            attrs_dict = self._attrs

        if attrs_dict[field].__class__.__name__ == cf_class.__name__:
            return attrs_dict[field]["path"]
        else:
            # perhaps its a field from the original root attributes
            if field in original_keys:
                return "/" + field

            # perhaps its a custom field, search in all CustomFieldSets for
            # this type of entity
            path_from_sets = self.__search_field_in_cfsets(field)
            if path_from_sets:
                return path_from_sets

            # not a CF, not an existing root property, it has no path,
            # try defaulting to root attribute, perhaps entity didn't
            # had it and now we need it to have it
            return "/" + field

    @classmethod
    def __build_object(
        cls,
        connector,
        resp,
        attrs,
        tzattrs,
        get_entities=False,
        detailsLevel="BASIC",
        debug=False,
    ):
        """Builds an instance of an Entity object

        Args:
          cls (obj): the object to build
          connector(obj): connector object to Mambu
          resp (bytes): the raw json that originates the object
          attrs (dict): the dict with the values to build the object
          tzattrs (dict): the dict with TZ data for datetimes in attrs
          get_entities (bool): should MambuPy automatically instantiate other
                               MambuPy entities found inside the built entity?
          detailsLevel (str): "BASIC" or "FULL"
          debug (bool): print debugging info
        """
        instance = cls.__call__(connector=connector)
        instance._resp = resp
        instance._attrs = attrs
        instance._tzattrs = tzattrs
        instance._convertDict2Attrs()
        instance._extractCustomFields()
        instance._extractVOs(get_entities=get_entities, debug=debug)
        entities = copy.deepcopy(instance._entities)
        if get_entities:
            instance._assignEntObjs(entities, detailsLevel, get_entities, debug=debug)
        instance._detailsLevel = detailsLevel

        return instance

    @classmethod
    def __get_several_args(cls, args, get_entities=False, debug=False):
        """Processes `MambuPy.api.entities.get_several` arguments.

        Args:
          cls (obj): the object for which the args are built
          args (dict): optional arguments received by get_several method
        """
        if "prefix" in args and args["prefix"] is not None:
            prefix = args.pop("prefix")
        else:
            prefix = cls._prefix

        if "get_entities" in args and args["get_entities"] is not None:
            get_entities = args.pop("get_entities")

        if "debug" in args and args["debug"] is not None:
            debug = args.pop("debug")

        params = copy.copy(args)
        if "detailsLevel" not in params:
            params["detailsLevel"] = "BASIC"

        return prefix, get_entities, debug, params

    @classmethod
    def _get_several(cls, get_func, connector, **kwargs):
        """get several entities.

        Using certain mambu connector function and its particular arguments.

        arg limit has a hard limit imposed by Mambu, that only 1,000 registers
        can be retrieved, so a pagination must be done to retrieve more
        registries.

        A pagination algorithm (using the offset and the 1,000 limitation) is
        applied here so that limit may be higher than 1,000 and _get_several
        will get a all the registers from Mambu in several requests.

        Args:
          get_func (function): mambu request function that returns several
                               entities (json [])
          connector (obj): connector object to Mambu
          kwargs (dict): keyword arguments to pass on to get_func as arguments

            - prefix (str): prefix for connections to Mambu

          kwargs (dict): keyword arguments to pass on to get_func as params

            - detailsLevel (str): "BASIC" or "FULL"
            - offset (int): >= 0
            - limit (int): >= 0 If limit=0 or None, the algorithm will retrieve
                           EVERYTHING according to the given filters, using
                           several requests to that end.

          kwargs (dict): keyword arguments for this method

            - get_entities (bool): should MambuPy automatically instantiate
                                   other MambuPy entities found inside the
                                   retrieved entities?
            - debug (bool): print debugging info

        Returns:
          list of instances of an entity with data from Mambu, assembled from
          possibly several calls to get_func
        """
        init_t = time.time()

        (prefix, get_entities, debug, params) = cls.__get_several_args(kwargs)

        logger.debug("request several entities %s", cls.__name__)
        list_resp = get_func(prefix, **params)
        jsonresp = list(json.loads(list_resp.decode()))
        logger.debug("%s, %s retrieved", cls.__name__, len(jsonresp))

        elements = []
        for attr in jsonresp:
            # builds the Entity object
            elem = cls.__build_object(
                connector=connector,
                resp=json.dumps(attr).encode(),
                attrs=attr,
                tzattrs=copy.deepcopy(attr),
                get_entities=get_entities,
                detailsLevel=params["detailsLevel"],
                debug=debug,
            )
            elements.append(elem)

        fin_t = time.time()
        interval = fin_t - init_t
        hours, remainder = divmod(interval, 3600)
        minutes = remainder // 60
        seconds = round(remainder - (minutes * 60), 2)
        if debug:
            if elements:
                print("{}-{}".format(
                    elements[0].__class__.__name__,
                    elements[0]["id"],
                ))
            print(
                "({}) {}:{}:{}".format(
                    len(elements),
                    int(hours),
                    int(minutes),
                    seconds,
                )
            )

        return elements

    @classmethod
    def get(cls, entid, detailsLevel="BASIC", get_entities=False, **kwargs):
        """get, a single entity, identified by its entid.

        Args:
          entid (str): ID for the entity
          detailsLevel (str BASIC/FULL): ask for extra details or not
          get_entities (bool): should MambuPy automatically instantiate other
                               MambuPy entities found inside the retrieved
                               entity?
          kwargs (dict): keyword arguments for this method.
                         May include a user, pwd and url to connect to Mambu.

            - debug (bool): print debugging info

        Returns:
          instance of an entity with data from Mambu
        """
        init_t = time.time()

        if "debug" in kwargs and kwargs["debug"]:
            debug = True
        else:
            debug = False

        connector = MambuConnectorREST(**kwargs)
        logger.debug("request entity %s %s", cls.__name__, entid)
        resp = connector.mambu_get(entid, prefix=cls._prefix, detailsLevel=detailsLevel)

        # builds the Entity object
        instance = cls.__build_object(
            connector=connector,
            resp=resp,
            attrs=dict(json.loads(resp.decode())),
            tzattrs=dict(json.loads(resp.decode())),
            get_entities=get_entities,
            detailsLevel=detailsLevel,
            debug=debug,
        )

        fin_t = time.time()
        interval = fin_t - init_t
        hours, remainder = divmod(interval, 3600)
        minutes = remainder // 60
        seconds = round(remainder - (minutes * 60), 2)
        if debug:
            print(
                "{}-{} {}:{}:{}".format(
                    instance.__class__.__name__,
                    instance["id"],
                    int(hours),
                    int(minutes),
                    seconds,
                )
            )

        return instance

    def refresh(self, detailsLevel="", **kwargs):
        """get again this single entity, identified by its entid.

        Updates _attrs with responded data. Loses any change on _attrs that
        overlaps with anything from Mambu. Leaves alone any other properties
        that don't come in the response.

        Args:
          detailsLevel (str BASIC/FULL): ask for extra details or not
          kwargs (dict): keyword arguments for this method.
                         May include a user, pwd and url to connect to Mambu.
        """
        if not detailsLevel:
            detailsLevel = self._detailsLevel
        self._resp = self._connector.mambu_get(
            self.id, prefix=self._prefix, detailsLevel=detailsLevel
        )

        self._attrs.update(dict(json.loads(self._resp.decode())))
        self._tzattrs = dict(json.loads(self._resp.decode()))
        self._convertDict2Attrs()
        self._extractCustomFields()
        self._extractVOs()
        self._detailsLevel = detailsLevel

    @classmethod
    def get_all(
        cls,
        filters=None,
        offset=None,
        limit=None,
        paginationDetails="OFF",
        detailsLevel="BASIC",
        sortBy=None,
        **kwargs
    ):
        """get_all, several entities, filtering allowed

        Args:
          filters (dict): key-value filters, dependes on each entity
                          (keys must be one of the _filter_keys)
          offset (int): pagination, index to start searching
          limit (int): pagination, number of elements to retrieve
          paginationDetails (str ON/OFF): ask for details on pagination
          detailsLevel (str BASIC/FULL): ask for extra details or not
          sortBy (str): ``field1:ASC,field2:DESC``, sorting criteria for
                        results (fields must be one of the _sortBy_fields)
          kwargs (dict): extra parameters that a specific entity may receive in
                         its get_all method. May include a user, pwd and url to
                         connect to Mambu.

        Returns:
          list of instances of an entity with data from Mambu
        """
        if filters:
            for filter_k in [fk for fk in filters.keys() if fk not in cls._filter_keys]:
                raise MambuPyError(
                    "key {} not in allowed _filterkeys: {}".format(
                        filter_k, cls._filter_keys
                    )
                )

        if sortBy:
            for sort in sortBy.split(","):
                for num, part in [
                    (n, p)
                    for n, p in enumerate(sort.split(":"))
                    if (n == 0 and p not in cls._sortBy_fields)
                ]:
                    raise MambuPyError(
                        "field {} not in allowed _sortBy_fields: {}".format(
                            part, cls._sortBy_fields
                        )
                    )

        params = {
            "filters": filters,
            "offset": offset,
            "limit": limit,
            "paginationDetails": paginationDetails,
            "detailsLevel": detailsLevel,
            "sortBy": sortBy,
        }
        if kwargs:
            params.update(kwargs)

        connector = MambuConnectorREST(**kwargs)
        return cls._get_several(connector.mambu_get_all, connector, **params)


class MambuEntityWritable(MambuStruct, MambuWritable):
    """A Mambu object with writing capabilities."""

    def update(self, **kwargs):
        """updates a mambu entity

        Uses the current values of the _attrs to send to Mambu.
        Pre-requires that CustomFields are updated previously.
        Post-requires that CustomFields are extracted again.

        kwargs (dict): keyword arguments for this method.
                       May include a user, pwd and url to connect to Mambu.
        """
        self._updateVOs()
        self._updateCustomFields()
        self._serializeFields()
        try:
            self._connector.mambu_update(
                self.id, self._prefix, copy.deepcopy(self._attrs), **kwargs
            )
            # should I refresh _attrs? (either get request from Mambu or using the response)
        except MambuError:
            raise
        finally:
            self._convertDict2Attrs()
            self._extractCustomFields()
            self._extractVOs()

    def create(self, **kwargs):
        """creates a mambu entity

        Uses the current values of the _attrs to send to Mambu.
        Pre-requires that CustomFields are updated previously.
        Post-requires that CustomFields are extracted again.

        kwargs (dict): keyword arguments for this method.
                       May include a user, pwd and url to connect to Mambu.
        """
        self._updateVOs()
        self._updateCustomFields()
        self._serializeFields()
        try:
            self._resp = self._connector.mambu_create(
                self._prefix, copy.deepcopy(self._attrs), **kwargs
            )
            self._attrs.update(dict(json.loads(self._resp.decode())))
            self._tzattrs = dict(json.loads(self._resp.decode()))
            self._detailsLevel = "FULL"
        except MambuError:
            raise
        finally:
            self._convertDict2Attrs()
            self._extractCustomFields()
            self._extractVOs()

    def __patch_op(self, operation, attrs, original_keys, field, cf_class):
        """Returns a patch operation tuple.

        Args:
          operation (str): One of ADD, REPLACE or REMOVE (MOVE not supproted yet)
          attrs (dict): the attrs structure holding the field
          original_keys (list): original keys for the entity attrs
          field (str): the field for which to apply a patch operation
          cf_class (obj): the class used for custom field VOs

        Returns:
          (tuple): according to the PATCH operation
          None: in case the field is not a valid attribute to patch
        """
        path = self._extract_field_path(field, attrs, original_keys, cf_class)
        if not path:
            return
        if operation == "REMOVE":
            return (operation, path)
        try:
            val = attrs[field]["value"]
        except (TypeError, KeyError):
            val = attrs[field]
        return (operation, path, val)

    def __patch_field_op(self, field, original_attrs):
        """Returns a valid ADD or REPLACE operation tuple for a field in attrs."""
        if field in self._attrs.keys() and field not in original_attrs.keys():
            # op says if this is an actual field or CF to
            # patch. If it's not, DO NOHTING
            op = self.__patch_op(
                "ADD", self._attrs, original_attrs.keys(), field, self._cf_class
            )
            if op:
                return op
        elif field in self._attrs.keys() and field in original_attrs.keys():
            return self.__patch_op(
                "REPLACE", self._attrs, original_attrs.keys(), field, self._cf_class
            )
        else:
            raise MambuPyError("Unrecognizable field {} for patching".format(field))

    def patch(self, fields=None, autodetect_remove=False, **kwargs):
        """patches a mambu entity

        Allows patching of parts of the entity up to Mambu.

        fields is a list of the keys in the _attrs that will be sent to Mambu

        autodetect automatically searches for deleted fields and patches a
        remove in Mambu.

        Pre-requires that CustomFields are updated previously.
        Post-requires that CustomFields are extracted again.

        Args:
          fields (list of str): list of ids of fields to explicitly patch
          autodetect_remove (bool): False: if deleted fields, don't remove them
                                    True: if delete field, remove them

        Autodetect operation, for any field (every field if fields param is
        None):
          * ADD: in attrs, but not in resp
          * REPLACE: in attrs, and in resp
          * REMOVE: not in attrs, but in resp
          * MOVE: not yet implemented (how to autodetect? request needs 'from' element)

        Raises:
          `MambuPyError`: if field not in attrrs, and not in resp
        """
        # customfields:
        # standard, as other fields (path includes CFset)
        #   REPLACE and REMOVE includes CF in path after CFset
        # grouped:
        #  ADD value is a list of CFs (they queue at end of group)
        #  REPLACE and REMOVE need index in path after CFset,
        #    CF at last replaces/removes just some field in the CFset at the index
        #  remove may have just index in path (removes entire CF)
        #    just CFset in path (removes entire group)
        #  you cannot add or replace entire CF (like by using just the index)

        if not fields:
            fields = []
        try:
            # build fields_ops param with what detected using previous rules
            # strings: (OP, PATH) (remove) or (OP, PATH, VALUE) (all else)
            fields_ops = []
            original_attrs = dict(json.loads(self._resp.decode()))
            self._extractCustomFields(original_attrs)
            self._updateVOs()
            for field in fields:
                field_op = self.__patch_field_op(field, original_attrs)
                if field_op:
                    fields_ops.append(field_op)
                else:
                    raise MambuPyError("You cannot patch {} field".format(field))
            if autodetect_remove:
                for attr in [
                    att for att in original_attrs.keys() if att not in self._attrs.keys()
                ]:
                    fields_ops.append(
                        self.__patch_op(
                            "REMOVE", original_attrs, original_attrs, attr, self._cf_class
                        )
                    )

            self._updateCustomFields()
            self._serializeFields()
            if fields_ops:
                self._connector.mambu_patch(self.id, self._prefix, fields_ops, **kwargs)
            # should I refresh _attrs? (needs get request from Mambu)
        except MambuError:
            raise
        finally:
            self._convertDict2Attrs()
            self._extractCustomFields()
            self._extractVOs()

    def _delete_for_creation(self):
        """Deletes extra fields from Mambu unusable for entity
        creation.

        Implement on entities which require further cleaning.
        """

    def delete(self, **kwargs):
        """deletes a mambu entity.

        Leaves the object in a supposedly state where you can create
        it again. Depends on required custom fields to be present in
        the entity at the moment of deletion.

        TODO: implement some way to ensure schemas are accomplished to
        make sure we can run creation after deletion.
        """
        try:
            self._connector.mambu_delete(self.id, self._prefix, **kwargs)
        except MambuError:
            raise

        # cleaning...
        for key in ["id", "encodedKey"]:
            try:
                del self._attrs[key]
            except KeyError:
                pass
        self._delete_for_creation()


class MambuEntitySearchable(MambuStruct, MambuSearchable):
    """A Mambu object with searching capabilities."""

    @classmethod
    def search(
        cls,
        filterCriteria=None,
        sortingCriteria=None,
        offset=None,
        limit=None,
        paginationDetails="OFF",
        detailsLevel="BASIC",
        **kwargs
    ):
        """search, several entities, filtering criteria allowed

        Args:
          filterCriteria (list of dicts): fields according to
                              each entity FilterCriteria schema
          sortingCriteria (dict): fields according to
                            each entity SortingCriteria
          offset (int): pagination, index to start searching
          limit (int): pagination, number of elements to retrieve
          paginationDetails (str ON/OFF): ask for details on pagination
          detailsLevel (str BASIC/FULL): ask for extra details or not

        Returns:
          list of instances of an entity with data from Mambu
        """
        params = {
            "filterCriteria": filterCriteria,
            "sortingCriteria": sortingCriteria,
            "offset": offset,
            "limit": limit,
            "paginationDetails": paginationDetails,
            "detailsLevel": detailsLevel,
        }
        if kwargs:
            params.update(kwargs)

        connector = MambuConnectorREST(**kwargs)
        return cls._get_several(connector.mambu_search, connector, **params)


class MambuEntityAttachable(MambuStruct, MambuAttachable):
    """A Mambu object with attaching capabilities."""

    _ownerType = ""
    """attachments owner type of this entity"""

    _attachments = {}
    """dict of attachments of an entity, key is the id"""

    def attach_document(self, filename, title="", notes="", **kwargs):
        """uploads an attachment to this entity

        _attachments dicitionary gets a new entry with the attached document.

        Args:
          filename (str): path and filename of file to upload as attachment
          title (str): name to assign to the attached file in Mambu
          notes (str): notes to associate to the attached file in Mambu

        Returns:
          Mambu's response with metadata of the attached document
        """
        response = self._connector.mambu_upload_document(
            owner_type=self._ownerType,
            entid=self.id,
            filename=filename,
            name=title,
            notes=notes,
        )

        doc = MambuDocument(**dict(json.loads(response.decode())))
        self._attachments[str(doc["id"])] = doc

        return response

    def get_attachments_metadata(
        self,
        offset=None,
        limit=None,
        paginationDetails="OFF",
    ):
        """Gets metadata for all the documents attached to an entity

        _attachments dicitionary is cleaned and set with attached documents.

        Args:
          offset (int): pagination, index to start searching
          limit (int): pagination, number of elements to retrieve
          paginationDetails (str ON/OFF): ask for details on pagination

        Returns:
          Mambu's response with metadata of the attached documents
        """
        self._attachments = {}

        response = self._connector.mambu_get_documents_metadata(
            entid=self.id,
            owner_type=self._ownerType,
            offset=offset,
            limit=limit,
            paginationDetails=paginationDetails,
        )

        metadata_list = json.loads(response.decode())

        for metadata in metadata_list:
            doc = MambuDocument(**metadata)
            self._attachments[str(doc["id"])] = doc

        return response

    def del_attachment(self, documentName=None, documentId=None):
        """deletes an attachment by its documentName.

        Args:
          documentName (str): optional, the name (title)
                              of the document to be deleted
          documentid (str): optional, the id of the document to be deleted

        Raises:
          `MambuPyError`: if the documentId and the documentName is not an
                          attachment of the entity
        """
        if not documentId and not documentName:
            raise MambuPyError("You must provide a documentId or a documentName")
        docid = None
        docbyname = None
        if documentName:
            try:
                docbyname = [
                    attach
                    for attach in self._attachments.values()
                    if attach["name"] == documentName
                ][0]
                docid = docbyname["id"]
            except IndexError:
                raise MambuPyError(
                    "Document name '{}' is not an attachment of {}".format(
                        documentName, repr(self)
                    )
                )
        if documentId:
            try:
                self._attachments[documentId]
                if docid:
                    if docid != documentId:
                        raise MambuPyError(
                            "Document with name '{}' does not has id '{}'".format(
                                documentName, documentId
                            )
                        )
                else:
                    docid = documentId
            except KeyError:
                raise MambuPyError(
                    "Document id '{}' is not an attachment of {}".format(
                        documentId, repr(self)
                    )
                )

        self._connector.mambu_delete_document(docid)
        self._attachments.pop(docid)


class MambuEntityCommentable(MambuStruct, MambuCommentable):
    """A Mambu object with commenting capabilities."""

    _comments = []
    """list of comments of an entity"""

    def get_comments(self, offset=None, limit=None, paginationDetails="OFF"):
        """Gets comments for this entity

        _comments list is cleaned and set with retrieved comments

        Args:
          offset (int): pagination, index to start searching
          limit (int): pagination, number of elements to retrieve
          paginationDetails (str ON/OFF): ask for details on pagination

        Returns:
          Mambu's response with retrieved comments
        """
        self._comments = []
        response = self._connector.mambu_get_comments(
            owner_id=self.id,
            owner_type=self._ownerType,
            offset=offset,
            limit=limit,
            paginationDetails=paginationDetails,
        )

        comments_list = json.loads(response.decode())

        for comment in comments_list:
            comm = MambuComment(**comment)
            self._comments.append(comm)

        return comments_list

    def comment(self, comment):
        """Comments this entity

        _comments list gets a new entry with comment.

        Args:
          comment (str): the text of the comment

        Returns:
          Mambu's response with metadata of the posted comment
        """
        response = self._connector.mambu_comment(
            owner_id=self.id, owner_type=self._ownerType, text=comment
        )

        comment = MambuComment(**dict(json.loads(response.decode())))
        self._comments.insert(0, comment)

        return response


class MambuEntityCF(MambuValueObject):
    """A Mambu CustomField obtained via an Entity.

    This is NOT a CustomField obtained through it's own endpoint, those go in
    another separate class.

    Here you just have a class to manage custom field values living inside some
    Mambu entity.
    """

    def __init__(self, value, path="", typecf="STANDARD", mcf=None):
        if typecf not in ["STANDARD", "GROUPED"]:
            raise MambuPyError("invalid CustomField type!")
        self._attrs = {"value": value, "path": path, "type": typecf, "mcf": mcf}
        self._cf_class = GenericClass

    def get_mcf(self):
        """Instance the MambuCustomField (MCF) of this entityCF.

        The MCF is set in the mcf property of this object.

        If this entityCF is a list, the mcf is set to a dictionary of
        field-MCFields.
        """
        mcf_mod = import_module("MambuPy.api.mambucustomfield")
        if self.mcf:
            return
        if isinstance(self.value, list) and len(self.value) > 0:
            self.mcf = {}
            for item in self.value:
                self.mcf["_index"] = None
                for key in [
                    k for k in item.keys() if k not in self.mcf and k != "_index"
                ]:
                    try:
                        self.mcf[key] = mcf_mod.MambuCustomField.get(key)
                    except MambuError:
                        self.mcf[key] = None
            return
        self.mcf = mcf_mod.MambuCustomField.get(self.path.split("/")[-1])


class MambuInstallment(MambuStruct):
    """Loan Account Installment (aka Repayment).

    Defaults dueDate to UTC since timezone is not relevant for this entity.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._as_utc = True

    def __repr__(self):
        """repr tells the class name, the number, the state and the dueDate"""
        return self.__class__.__name__ + " - #{}, {}, {}".format(
            self._attrs["number"],
            self._attrs["state"],
            self._attrs["dueDate"].strftime("%Y-%m-%d"),
        )


class MambuEntityOwnable(MambuStruct, MambuOwnable):
    """An entity which allows to be 'owned' by another.

    An owned entity has an 'accountHolderKey' and 'accountHolderType'
    fields.

    Because of that, you may call get_accountHolder on the owned
    entity to instantiate the MambuEntity who owns it.
    """

    def _assignEntObjs(
        self, entities=None, detailsLevel="BASIC", get_entities=False, debug=False
    ):
        """Overwrites `MambuPy.api.mambustruct._assignEntObjs` for MambuLoan

        Determines the type of account holder and instantiates accordingly
        """
        if entities is None:
            entities = self._entities

        try:
            accountholder_index = entities.index(
                ("accountHolderKey", "", "accountHolder")
            )
        except ValueError:
            accountholder_index = None

        if accountholder_index is not None and self.has_key("accountHolderKey"):
            if self.accountHolderType == "CLIENT":
                entities[accountholder_index] = (
                    "accountHolderKey",
                    "mambuclient.MambuClient",
                    "accountHolder",
                )
            elif self.accountHolderType == "GROUP":
                entities[accountholder_index] = (
                    "accountHolderKey",
                    "mambugroup.MambuGroup",
                    "accountHolder",
                )

        return super()._assignEntObjs(
            entities, detailsLevel=detailsLevel, get_entities=get_entities, debug=debug
        )



"""MambuTransaction entity: a MambuEntity struct for credit Transactions.

.. autosummary::
   :nosignatures:
   :toctree: _autosummary
"""

from .entities import MambuEntity, MambuEntitySearchable


class MambuTransaction(MambuEntity, MambuEntitySearchable):
    """MambuTransaction entity"""

    _prefix = "loans/transactions"
    """prefix constant for connections to Mambu"""

    _filter_keys = []
    """allowed filters for get_all filtering"""

    _sortBy_fields = []
    """allowed fields for get_all sorting"""

    @classmethod
    def get_all(
        cls,
        loanAccountId,
        offset=None,
        limit=None,
        paginationDetails="OFF",
        detailsLevel="BASIC",
        **kwargs,
    ):
        """get_all, several transactions, filtering allowed

        Args:
          loanAccountId (str): the id or encodedkey of the loan account
          offset (int): pagination, index to start searching
          limit (int): pagination, number of elements to retrieve
          paginationDetails (str ON/OFF): ask for details on pagination
          detailsLevel (str BASIC/FULL): ask for extra details or not

        Returns:
          list of instances of a transaction with data from Mambu
        """
        kwargs.update({"prefix": "loans/{}/transactions".format(loanAccountId)})
        return super().get_all(
            filters=None,
            offset=offset,
            limit=limit,
            paginationDetails=paginationDetails,
            detailsLevel=detailsLevel,
            sortBy=None,
            **kwargs,
        )

    def adjust(self, notes):
        """Request to adjust a loan transaction.

        Args:
          notes (str): notes to attach to the adjusting transaction.
        """
        self._connector.mambu_loantransaction_adjust(self.id, notes)
        self.refresh()



"""Schema tables for Mambu Loan Accounts.

.. autosummary::
   :nosignatures:
   :toctree: _autosummary

.. todo:: this are just very basic schemas for loan tables. A lot of fields
          are missing.
"""

from sqlalchemy import Column, DateTime, ForeignKey, Integer, Numeric, String
from sqlalchemy.orm import relationship

from . import schema_orm as orm
from .schema_branches import Branch
from .schema_centres import Centre
from .schema_users import User

dbname = orm.dbname
session = orm.session
Base = orm.Base


class LoanProduct(Base):
    """LoanProduct table."""

    __tablename__ = "loanproduct"
    __table_args__ = {"schema": dbname, "keep_existing": True}

    # Columns
    encodedKey = Column(String, primary_key=True)
    id = Column(String, index=True, unique=True)
    productName = Column(String)  # redundant with same-as-RESTAPI-case
    activated = Column(Integer)
    loans = relationship("LoanAccount", back_populates="product")

    def __repr__(self):
        return "<LoanProduct(id=%s, name=%s)>" % (self.id, self.productName)


class DisbursementDetails(Base):
    """DisbursementDetails table."""

    __tablename__ = "disbursementdetails"
    __table_args__ = {"schema": dbname, "keep_existing": True}

    # Columns
    encodedKey = Column(String, primary_key=True)
    expectedDisbursementDate = Column(DateTime)
    disbursementDate = Column(DateTime)
    firstRepaymentDate = Column(DateTime)

    def __repr__(self):
        return "<DisbursementDetails(disbursementDate=%s)>" % (self.disbursementDate)


class LoanAccount(Base):
    """LoanAccount table."""

    __tablename__ = "loanaccount"
    __table_args__ = {"schema": dbname, "keep_existing": True}

    # Columns
    encodedKey = Column(String, primary_key=True)
    id = Column(String, index=True, unique=True)
    accountState = Column(String)
    accountSubstate = Column(String)  # not in API
    loanAmount = Column(Numeric(50, 10))
    notes = Column(String)
    principalBalance = Column(Numeric(50, 10))
    principalPaid = Column(Numeric(50, 10))
    principalDue = Column(Numeric(50, 10))
    interestBalance = Column(Numeric(50, 10))
    interestPaid = Column(Numeric(50, 10))
    interestDue = Column(Numeric(50, 10))
    interestRate = Column(Numeric(50, 10))
    interestCalculationMethod = Column(String)
    interestBalanceCalculationMethod = Column(String)
    repaymentInstallments = Column(Integer)
    repaymentPeriodUnit = Column(String)
    repaymentPeriodCount = Column(Integer)
    accountHolderType = Column(String)
    feesBalance = Column(Numeric(50, 10))
    feesPaid = Column(Numeric(50, 10))
    feesDue = Column(Numeric(50, 10))
    penaltyBalance = Column(Numeric(50, 10))
    penaltyPaid = Column(Numeric(50, 10))
    penaltyDue = Column(Numeric(50, 10))
    creationDate = Column(DateTime)
    approvedDate = Column(DateTime)
    closedDate = Column(DateTime)
    rescheduledAccountKey = Column(String)
    lastSetToArrearsDate = Column(DateTime)  # not in API

    # Relationships
    productTypeKey = Column(String, ForeignKey(LoanProduct.encodedKey))
    product = relationship("LoanProduct", back_populates="loans")
    disbursementDetailsKey = Column(String, ForeignKey(DisbursementDetails.encodedKey))
    disbursementDetails = relationship("DisbursementDetails")
    assignedBranchKey = Column(String, ForeignKey(Branch.encodedKey))
    branch = relationship("Branch", back_populates="loans")
    assignedCentreKey = Column(String, ForeignKey(Centre.encodedKey))
    centre = relationship("Centre", back_populates="loans")
    assignedUserKey = Column(String, ForeignKey(User.encodedKey))
    user = relationship("User", back_populates="loans")
    accountHolderKey = Column(String)
    holder_group = relationship(
        "Group",
        back_populates="loans",
        foreign_keys="LoanAccount.accountHolderKey",
        primaryjoin="LoanAccount.accountHolderKey == Group.encodedKey",
    )
    holder_client = relationship(
        "Client",
        back_populates="loans",
        foreign_keys="LoanAccount.accountHolderKey",
        primaryjoin="LoanAccount.accountHolderKey == Client.encodedKey",
    )
    customInformation = relationship(
        "CustomFieldValue",
        back_populates="loan",
        foreign_keys="CustomFieldValue.parentKey",
        primaryjoin="CustomFieldValue.parentKey == LoanAccount.encodedKey",
    )
    tasks = relationship(
        "Task",
        back_populates="link_loan",
        foreign_keys="Task.taskLinkKey",
        primaryjoin="Task.taskLinkKey == LoanAccount.encodedKey",
    )
    activities = relationship("Activity", back_populates="loan")
    repayments = relationship("Repayment", back_populates="account")
    transactions = relationship("LoanTransaction", back_populates="account")

    def __repr__(self):
        return "<LoanAccount(id=%s, accountState=%s)>" % (self.id, self.accountState)


class Repayment(Base):
    """Repayment table."""

    __tablename__ = "repayment"
    __table_args__ = {"schema": dbname, "keep_existing": True}

    # Columns
    encodedKey = Column(String, primary_key=True)
    dueDate = Column(DateTime, index=True)
    principalDue = Column(Numeric(50, 10))
    principalPaid = Column(Numeric(50, 10))
    interestDue = Column(Numeric(50, 10))
    interestPaid = Column(Numeric(50, 10))
    feesDue = Column(Numeric(50, 10))
    feesPaid = Column(Numeric(50, 10))
    penaltyDue = Column(Numeric(50, 10))
    penaltyPaid = Column(Numeric(50, 10))
    repaidDate = Column(DateTime)
    lastPaidDate = Column(DateTime)
    state = Column(String, index=True)

    # Relationships
    parentAccountKey = Column(String, ForeignKey(LoanAccount.encodedKey))
    account = relationship(
        "LoanAccount", back_populates="repayments", order_by="Repayment.dueDate"
    )

    def __repr__(self):
        return "<Repayment(dueDate=%s, state=%s,\naccount=%s)>" % (
            self.dueDate.strftime("%Y%m%d"),
            self.state,
            self.account,
        )


class TransactionChannel(Base):
    """TransactionChannel table."""

    __tablename__ = "transactionchannel"
    __table_args__ = {"schema": dbname, "keep_existing": True}

    # Columns
    encodedKey = Column(String, primary_key=True)
    id = Column(String)
    name = Column(String)

    def __repr__(self):
        return "<TransactionChannel(name=%s)>" % (self.name)


class TransactionDetails(Base):
    """TransactionDetails table."""

    __tablename__ = "transactiondetails"
    __table_args__ = {"schema": dbname, "keep_existing": True}

    # Columns
    encodedKey = Column(String, primary_key=True)

    # Relationships
    transactionChannelKey = Column(String, ForeignKey(TransactionChannel.encodedKey))
    transactionChannel = relationship("TransactionChannel")

    def __repr__(self):
        return "<TransactionDetails(TransactionChannel=%s)>" % (self.transactionChannel)


class LoanTransaction(Base):
    """LoanTransaction table."""

    __tablename__ = "loantransaction"
    __table_args__ = {"schema": dbname, "keep_existing": True}

    # Columns
    encodedKey = Column(String, primary_key=True)
    transactionId = Column(Integer, index=True)
    creationDate = Column(DateTime)
    entryDate = Column(DateTime)
    principalAmount = Column(Numeric(50, 10))
    interestAmount = Column(Numeric(50, 10))
    feesAmount = Column(Numeric(50, 10))
    penaltyAmount = Column(Numeric(50, 10))
    reversalTransactionKey = Column(String)
    amount = Column(Numeric(50, 10))
    type = Column(String)
    comment = Column(String)

    # Relationships
    parentAccountKey = Column(String, ForeignKey(LoanAccount.encodedKey))
    account = relationship(
        "LoanAccount",
        back_populates="transactions",
        order_by="LoanTransaction.transactionId",
    )
    details_encodedkey_oid = Column(String, ForeignKey(TransactionDetails.encodedKey))
    transactionDetails = relationship("TransactionDetails")

    def __repr__(self):
        return "<LoanTransaction(transactionId=%s, amount=%s, creationDate=%s, entryDate=%s, type=%s, comment='%s',\
 reversed=%s\naccount=%s)>" % (
            self.transactionId,
            self.amount,
            self.creationDate.strftime("%Y%m%d"),
            self.entryDate.strftime("%Y%m%d"),
            self.type,
            self.comment,
            "Yes" if self.reversalTransactionKey else "No",
            self.account,
        )



"""Mambu Value Objects

.. autosummary::
   :nosignatures:
   :toctree: _autosummary
"""

from .mambustruct import MambuStruct


class MambuValueObject(MambuStruct):
    """A Mambu object with some schema but that you won't interact directly
    with in Mambu web, but through some entity."""


class MambuDocument(MambuValueObject):
    """Attached document"""


class MambuAddress(MambuValueObject):
    """Address"""


class MambuIDDocument(MambuValueObject):
    """ID Document"""

    _ownerType = "ID_DOCUMENT"
    """owner type of this entity"""


class MambuComment(MambuValueObject):
    """Comment"""


class MambuDisbursementDetails(MambuValueObject):
    """Disbursement Details"""


class MambuUserRole(MambuValueObject):
    """User Role"""


class MambuGroupMember(MambuValueObject):
    """Group member"""

    _vos = [("roles", "MambuGroupRole")]
    """2-tuples of elements and Value Objects"""

    _entities = [("clientKey", "mambuclient.MambuClient", "client")]
    """3-tuples of elements and Mambu Entities"""


class MambuGroupRole(MambuValueObject):
    """Group member role"""


class MambuDisbursementLoanTransactionInput(MambuValueObject):
    """Disbursment Loan Transaction body"""

    _schema_fields = [
        "amount",
        "bookingDate",
        "externalId",
        "firstRepaymentDate",
        "notes",
        "originalCurrencyCode",
        "shiftAdjustableInterestPeriods",
        "valueDate",
    ]
    """List of schema fields for a loan disbursement transaction."""


class MambuRepaymentLoanTransactionInput(MambuValueObject):
    """Repayment Loan Transaction body"""

    _schema_fields = [
        "amount",
        "bookingDate",
        "externalId",
        "installmentEncodedKey",
        "notes",
        "originalCurrencyCode",
        "prepaymentRecalculationMethod",
        "valueDate",
        "transactionDetails",
    ]
    """List of schema fields for a loan repayment transaction."""


class MambuLoanTransactionDetailsInput(MambuValueObject):
    """Loan Transaction Details body"""

    _schema_fields = [
        "transactionChannelId",
        "transactionChannelKey",
    ]
    """List of schema fields for a loan repayment transaction."""


class MambuFeeLoanTransactionInput(MambuValueObject):
    """Repayment Loan Transaction body"""

    _schema_fields = [
        "amount",
        "bookingDate",
        "externalId",
        "fisrtRepaymentDate",
        "installmentNumber",
        "notes",
        "predefinedFeeKey",
        "valueDate",
    ]
    """List of schema fields for a loan fee transaction."""



"""Mambu utilites module.

.. autosummary::
   :nosignatures:
   :toctree: _autosummary

Exceptions, some API return codes, utility functions, a lot of urlfuncs
(see MambuStruct.__init__ and .connect pydocs for more info on this)

.. warning::

   Secutiry WARNING: Imports the configurations from mambuconfig. It surely
   is a bad idea. Got to improve this!

.. todo:: status API V2: fullDetails are not compatible

.. todo:: status API V2: pagination managements needs review

.. todo:: status API V2: almost al GET operations seems
          compatible. PATCH/POST/DELETE operations needs review

.. todo:: status API V2: testing of EVERYTHING is required """

import logging
import logging.config as logging_config
import os
from codecs import open as copen

import yaml

from .mambuconfig import apipagination, apipwd, apiurl, apiuser, loggingconf, activate_request_session_objects
from .mambugeturl import getmambuurl

import json
import sys
from datetime import datetime, timezone

API_RETURN_CODES = {
    "SUCCESS": 0,
    "INVALID_PARAMETERS": 4,
    "INVALID_LOAN_ACCOUNT_ID": 100,
    "INVALID_ACCOUNT_STATE": 105,
    "EXCESS_REPAYMENT_ERROR": 110,
    "INVALID_REPAYMENT_DATE_ERROR": 111,
    "INVALID_STATE_TRANSITION": 116,
    "INVALID_ACCOUNT_STATE_FOR_REPAYMENTS": 128,
    "INCONSISTENT_SCHEDULE_PRINCIPAL_DUE_WITH_LOAN_AMOUNT": 132,
}
"""
.. deprecated:: 0.8
  It's probably not useful. Besides, :any:`mambustruct.MambuStruct.connect()`
  method returns the code when an error occurs by default.
"""


OUT_OF_BOUNDS_PAGINATION_LIMIT_VALUE = int(apipagination)
"""Current maximum number of response elements Mambu returns"""


PAGINATIONDETAILS = ["ON", "OFF"]
"""paginationDetails options"""


DETAILSLEVEL = ["BASIC", "FULL"]
"""detailsLevel options"""

MAX_UPLOAD_SIZE = 52428800  # 50Mb
"""upload files maximum size"""

UPLOAD_FILENAME_INVALID_CHARS = "/><|:&?*[]#\*`"
"""invalid characters for an uploaded filename"""

ALLOWED_UPLOAD_MIMETYPES = [
    "JPEG",
    "PNG",
    "GIF",
    "BMP",
    "TIFF",
    "PDF",
    "XML",
    "DOC",
    "DOCX",
    "DOCM",
    "DOT",
    "DOTX",
    "DOTM",
    "XLS",
    "XLSX",
    "XLSB",
    "PPT",
    "PPTX",
    "ODT",
    "OTT",
    "FODT",
    "PDF",
    "XML",
    "TXT",
    "CSV",
    "PROPERTIES",
    "MSG",
    "TIF",
    "ZIP",
    "RTF",
    "XLSM",
    "ODS",
    "ODP",
    "EML",
    "EMLX",
    "HTML",
    "MHT",
    "MHTML",
    "XPS",
    "NUMBERS",
    "KEY",
    "PAGES",
    "YAML",
    "JSON",
    "JASPER",
    "JRXML",
]
"""allowed file types for uploads"""

SEARCH_OPERATORS = [
    "EQUALS",
    "EQUALS_CASE_SENSITIVE",
    "DIFFERENT_THAN",
    "MORE_THAN",
    "LESS_THAN",
    "BETWEEN",
    "ON",
    "AFTER",
    "BEFORE",
    "BEFORE_INCLUSIVE",
    "STARTS_WITH",
    "STARTS_WITH_CASE_SENSITIVE",
    "IN",
    "TODAY",
    "THIS_WEEK",
    "THIS_MONTH",
    "THIS_YEAR",
    "LAST_DAYS",
    "EMTPY",
    "NOT_EMPTY",
]
"""search operators"""


class MambuPyError(Exception):
    """Default exception"""


class MambuError(MambuPyError):
    """Default exception from Mambu"""


class MambuCommError(MambuError):
    """Thrown when communication issues with Mambu arise"""


### More utility functions follow ###
def date_format(field, formato=None, as_utc=False):
    """Converts a datetime field to a datetime using some specified format.

    What this really means is that, if specified format includes only for
    instance just year and month, day and further info gets ignored and the
    objects get a datetime with year and month, and day 1, hour 0, minute 0,
    etc.

    A useful format may be %Y%m%d, then the datetime objects effectively
    translates into date objects alone, with no relevant time information.

    When as_utc is True, the datetime object is converted to UTC timezone.

    PLEASE BE AWARE, that this may lose useful information for your datetimes
    from Mambu. Read this for why this may be a BAD idea:
    https://julien.danjou.info/blog/2015/python-and-timezones
    """
    from datetime import datetime

    if not formato:
        formato = "%Y-%m-%dT%H:%M:%S"

    field_with_tz = datetime.fromisoformat(field)

    if as_utc and field_with_tz.tzinfo is not None:
        # Convertir a UTC mientras aún tiene tzinfo
        field_as_dt = field_with_tz.astimezone(timezone.utc)
        field_as_dt = datetime.strptime(
            field_as_dt.strftime(formato),
            formato,
        )
    else:
        field_as_dt = datetime.strptime(
            field_with_tz.strftime(formato),
            formato,
        )

    return field_as_dt


def strip_tags(html):
    """Stripts HTML tags from text.

    Note fields on several Mambu entities come with additional HTML tags
    (they are rich text fields, I guess that's why). Sometimes they are
    useless, so stripping them is a good idea.
    """
    from html.parser import HTMLParser

    class MLStripper(HTMLParser):
        """Aux class for stripping HTML tags.

         fields on several Mambu entities come with additional HTML tags
        (they are rich text fields, I guess that's why). Sometimes they are
        useless, so stripping them is a good idea.
        """

        def __init__(self):
            super().__init__()
            self.reset()
            self.fed = []

        def handle_data(self, d):
            self.fed.append(d)

        def get_data(self):
            return "".join(self.fed)

    s = MLStripper()
    s.feed(html.replace("&nbsp;", " "))
    return s.get_data()


def strip_consecutive_repeated_char(s, ch):
    """Strip characters in a string which are consecutively repeated.

    Useful when in notes or some other free text fields on Mambu, users
    capture anything and a lot of capture errors not always detected by
    Mambu get through. You want some cleaning? this may be useful.

    This is a string processing function.
    """
    sdest = ""
    for i, c in enumerate(s):
        if i != 0 and s[i] == ch and s[i] == s[i - 1]:
            continue
        sdest += s[i]
    return sdest


from urllib import parse as urlparse


def iri_to_uri(iri):
    """Change an IRI (internationalized R) to an URI.

    Used at MambuStruct.connect() method for any requests done to Mambu.

    Perfect example of unicode getting in the way.

    Illustrative (I hope) example: I have Mambu usernames with special
    chars in them. When retrieving them and then trying to build a
    MambuUser object with them, I get a BIG problem because of the
    unicode chars there. Using this I solved the problem.
    """

    parts = urlparse.urlparse(iri)
    uri = [part.decode("utf8") for parti, part in enumerate(parts.encode("utf8"))]
    return urlparse.urlunparse(uri)


def encoded_dict(in_dict):
    """Encode every value of a dict to UTF-8.

    Useful for POSTing requests on the 'data' parameter of urlencode.
    """
    out_dict = {}
    for k, v in in_dict.items():
        out_dict[k] = v
    return out_dict


from time import sleep

import requests


def _backup_db_previous_prep(callback, logger, kwargs):
    list_ret = []
    try:
        retries = kwargs["retries"]
    except KeyError:
        retries = -1
    list_ret.append(retries)
    try:
        justbackup = kwargs["justbackup"]
    except KeyError:
        justbackup = False
    list_ret.append(justbackup)
    try:
        force_download_latest = bool(kwargs["force_download_latest"])
    except KeyError:
        force_download_latest = False
    list_ret.append(force_download_latest)
    try:
        headers = kwargs["headers"]
    except KeyError:
        headers = {
            "Accept": "application/vnd.mambu.v2+zip",
        }
    list_ret.append(headers)

    logger.info("Mambu DB Backup")

    user = kwargs.pop("user", apiuser)
    list_ret.append(user)

    pwd = kwargs.pop("pwd", apipwd)
    list_ret.append(pwd)

    data = {"callback": callback}
    list_ret.append(data)

    return list_ret


def _backup_db_request(justbackup, data, user, pwd, logger=None):
    try:
        if not justbackup:
            posturl = iri_to_uri(getmambuurl() + "database/backup")
            logger.info("open url: " + posturl)
            logger.info("data: " + str(data))
            resp = requests.post(
                posturl,
                data=json.dumps(data),
                headers={
                    "content-type": "application/json",
                    "Accept": "application/vnd.mambu.v2+json",
                },
                auth=(user, pwd),
            )
            logger.info(str(resp.content))
            logger.info(resp.request.url)
            logger.info(resp.request.body)
            logger.info(str([(k, v) for k, v in resp.request.headers.items() if k != "Authorization"]))
    except Exception as ex:
        mess = "Error requesting backup: %s" % repr(ex)
        logger.exception(mess)
        raise MambuError(mess)

    if not justbackup and resp.status_code != 202:
        mess = "Error posting request for backup: %s" % resp.content
        logger.error(mess)
        raise MambuCommError(mess)


def _backup_db_timeout_mechanism(
    justbackup, retries, bool_func, force_download_latest, logger=None
):
    value_to_latest = True
    while not justbackup and retries and not bool_func():
        logger.debug("waiting...")
        sleep(10)
        retries -= 1
        if retries < 0:
            retries = -1
    if not justbackup and not retries:
        mess = "Tired of waiting, giving up..."
        logger.warning(mess)
        if not force_download_latest:
            raise MambuError(mess)
        else:
            value_to_latest = False
    sleep(30)

    return value_to_latest


def _backup_db_request_download_backup(user, pwd, headers, logger=None):
    geturl = iri_to_uri(getmambuurl() + "database/backup/LATEST")
    logger.info("open url: " + geturl)
    resp = requests.get(geturl, auth=(user, pwd), headers=headers)

    if resp.status_code != 200:
        mess = "Error getting database backup: %s" % resp.content
        logger.error(mess)
        raise MambuCommError(mess)

    return resp


def _backup_db_post_processing(resp, output_fname, logger=None):
    logger.info("saving...")
    with open(output_fname, "wb") as fw:
        fw.write(resp.content)


def backup_db(callback, bool_func, output_fname, *args, **kwargs):
    """Backup Mambu Database via REST API.

    Makes two calls to Mambu API:

    - a POST to request a backup to be made

    - a GET, once the backup is ready, to download the latest backup

    * callback is a string to a callback URL Mambu will internally call
      when the backup is ready to download. You should have a webservice
      there to warn you when the backup is ready.

    * bool_func is a function you use against your own code to test if the
      said backup is ready. This function backup_db manages both the logic
      of the request of a backup and the downloading of it too, so
      bool_func allows you to have some way on your side to know when this
      function will download the backup.

      The thing is you have to build a webservice (for the callback)
      making some kind of flag turn that your bool_func will read and know
      when to say True, telling backup_db to begin the download of the
      backup.

    * output_fname the name of the file that will hold the downloaded
      backup. PLEASE MIND that Mambu sends a ZIP file here.

    * user, pwd and url allow you to change the Mambu permissions for
      the getmambuurl internally called here.

    * retries number of retries for bool_func or -1 for keep waiting.

    * just_backup bool if True, skip asking for backup, just download LATEST

    * force_download_latest boolean, True to force download even if no
      callback is called. False to throw error if callback isn't received
      after retries.

    * returns a dictionary with info about the download
        -latest     boolean flag, if the db downloaded was the latest or not

    .. todo:: status API V2: compatible
    """
    logger = setup_logging("mambupy.backup_db")
    # previous preparation
    (
        retries,
        justbackup,
        force_download_latest,
        headers,
        user,
        pwd,
        data,
    ) = _backup_db_previous_prep(callback, logger, kwargs)

    # POST to request Mambu to prepare backup of its own DB
    _backup_db_request(justbackup, data, user, pwd, logger)

    # wait & timeout mechanism
    data["latest"] = _backup_db_timeout_mechanism(
        justbackup, retries, bool_func, force_download_latest, logger
    )

    # GET request to download LATEST Mambu's DB backup
    resp = _backup_db_request_download_backup(user, pwd, headers, logger)

    # post-processing
    _backup_db_post_processing(resp, output_fname, logger)

    # no refactor
    logger.info("DONE!")

    return data


def setup_logging(
        loggername,
        default_level=logging.INFO,
) -> logging.Logger:
    """ Get a logger configured with a yaml file. """
    if os.path.exists(loggingconf):
        with copen(loggingconf, "r", "utf-8") as f:
            config = yaml.full_load(f.read())
        logging_config.dictConfig(config)
        loggr = logging.getLogger(loggername)
    else:
        loggr = logging.getLogger(loggername)
        loggr.setLevel(default_level)
        loggr.propagate = True
    return loggr


logger = setup_logging("mambupy")
"""Default logger for mambupy module"""



"""MambuCentre entity: a MambuEntity struct for Centres.

.. autosummary::
   :nosignatures:
   :toctree: _autosummary
"""

from .entities import MambuEntity, MambuEntityCommentable
from .interfaces import MambuOwner


class MambuCentre(MambuEntity, MambuEntityCommentable, MambuOwner):
    """MambuCentre entity"""

    _prefix = "centres"
    """prefix constant for connections to Mambu"""

    _filter_keys = [
        "branchId",
    ]
    """allowed filters for get_all filtering"""

    _sortBy_fields = [
        "creationDate",
        "lastModifiedDate",
        "id",
        "name",
    ]
    """allowed fields for get_all sorting"""

    _ownerType = "CENTRE"
    """owner type of this entity"""

    _vos = [("addresses", "MambuAddress")]
    """2-tuples of elements and Value Objects"""

    _entities = [
        ("assignedBranchKey", "mambubranch.MambuBranch", "assignedBranch"),
    ]
    """3-tuples of elements and Mambu Entities"""



from .mambuconfig import apiurl

def getmambuurl(url=apiurl, *args, **kwargs):
    """Basic Mambu URL function. Builds the Mambu basic request string.

    Receives a user, password and url arguments to connect to Mambu.
    Defaults to what mambuconfig has, but you may override it by sending
    'user', 'pwd' and 'url' parameters at the Mambu object instantiation
    (for instance: MambuLoan(user=myuser, pwd=mypwd, url=myurl,
    entid='12345', ...)

    .. todo:: status API V2: compatible
    """
    mambuurl = "https://" + url + "/api/"
    return mambuurl


###*** Behold, the urlfuncs magic follows ***###
# This things are the gateways for implementing parameters on requests done to
# the Mambu REST API.

def _get_parameters_url(dict_kwargs):
    """Get list with parameters for the url of getbranchesurl, getloansurl, getgroupurl, getclienturl,
    getactivitiesurl, getcentresurl, getrepaymentsurl, getsavingssurl, getsavingfundingrepaymentsurl,
    getsavingstransactionsurl, getsavingstransactionssearchurl, gettransactionchannelsurl, getgrouploansurl,
    gettransactionsurl, getclientloansurl, getuserurl, gettasksurl.
    """
    getparams_list = []

    for key in dict_kwargs.keys():
        if key == "fullDetails":
            if dict_kwargs[key] is True:
                getparams_list.append("fullDetails=true")
            else:
                getparams_list.append("fullDetails=false")
        elif key in ["user", "pwd"]:
            continue
        elif key in ["version"]:
            continue
        else:
            getparams_list.append(key + '=%s' % dict_kwargs[key])

    return getparams_list


def getbranchesurl(idbranch, *args, **kwargs):
    """Request Branches URL.

    If idbranch is set, you'll get a response adequate for a MambuBranch object.
    If not set, you'll get a response adequate for a MambuBranches object.
    See mambubranch module and pydoc for further information.

    Currently implemented filter parameters:
    * fullDetails
    * limit
    * offset

    See Mambu official developer documentation for further details, and
    info on parameters that may be implemented here in the future.

    .. todo:: status API V2: compatible
    """
    getparams = []
    if kwargs:
        getparams = _get_parameters_url(kwargs)

    branchidparam = "" if idbranch == "" else "/" + idbranch
    url = (
        getmambuurl(*args, **kwargs)
        + "branches"
        + branchidparam
        + ("" if len(getparams) == 0 else "?" + "&".join(getparams))
    )
    return url


def getcentresurl(idcentre, *args, **kwargs):
    """Request Centres URL.

    If idcentre is set, you'll get a response adequate for a MambuCentre object.
    If not set, you'll get a response adequate for a MambuCentres object.
    See mambucentre module and pydoc for further information.

    Currently implemented filter parameters:
    * fullDetails
    * limit
    * offset

    See Mambu official developer documentation for further details, and
    info on parameters that may be implemented here in the future.

    .. todo:: status API V2: compatible
    """
    getparams = []
    if kwargs:
        getparams = _get_parameters_url(kwargs)

    centreidparam = "" if idcentre == "" else "/" + idcentre
    url = (
        getmambuurl(*args, **kwargs)
        + "centres"
        + centreidparam
        + ("" if len(getparams) == 0 else "?" + "&".join(getparams))
    )
    return url


def getrepaymentsurl(idcred, *args, **kwargs):
    """Request loan Repayments URL.

    If idcred is set, you'll get a response adequate for a
    MambuRepayments object. There's a MambuRepayment object too, but
    you'll get a list first and each element of it will be automatically
    converted to a MambuRepayment object that you may use.

    If not set, you'll get a Jar Jar Binks object, or something quite
    strange and useless as JarJar. A MambuError must likely since I
    haven't needed it for anything but for repayments of one and just
    one loan account.

    See mamburepayment module and pydoc for further information.

    Currently implemented GET filter parameters:
        * fullDetails
        * limit
        * offset

    See Mambu official developer documentation for further details, and
    info on parameters that may be implemented here in the future.

    .. todo:: status API V2: NOT compatible
    """
    getparams = []
    if kwargs:
        getparams = _get_parameters_url(kwargs)

    url = (
        getmambuurl(*args, **kwargs)
        + "loans/"
        + idcred
        + "/repayments"
        + ("" if len(getparams) == 0 else "?" + "&".join(getparams))
    )
    return url


def getsavingssurl(idcred, *args, **kwargs):
    """Request loan savings URL.

    If idcred is set, you'll get a response adequate for a
    MambuSaving object.

    Currently implemented GET filter parameters:
        * fullDetails
        * limit
        * offset

    See Mambu official developer documentation for further details, and
    info on parameters that may be implemented here in the future.

    .. todo:: status API V2: NOT compatible . Apparently changing
              'savings' to 'deposits' may work
    """
    getparams = []
    if kwargs:
        getparams = _get_parameters_url(kwargs)

    url = (
        getmambuurl(*args, **kwargs)
        + "savings/"
        + idcred
        + ("" if len(getparams) == 0 else "?" + "&".join(getparams))
    )
    return url


def getsavingfundingrepaymentsurl(idcred, loan_id, *args, **kwargs):
    """Request loan saving Repayments URL.

    If idcred is set, you'll get a response adequate for a
    MambuSaving object.

    Currently implemented GET filter parameters:
        * fullDetails
        * limit
        * offset

    See Mambu official developer documentation for further details, and
    info on parameters that may be implemented here in the future.

    .. todo:: status API V2: NOT compatible
    """
    getparams = []
    if kwargs:
        getparams = _get_parameters_url(kwargs)

    url = (
        getmambuurl(*args, **kwargs)
        + "savings/"
        + idcred
        + "/funding/"
        + loan_id
        + "/repayments"
        + ("" if len(getparams) == 0 else "?" + "&".join(getparams))
    )
    return url


def getsavingstransactionsurl(idcred, *args, **kwargs):
    """Request saving transaction URL.

    If idcred is set, you'll get a response adequate for a
    MambuSavingTransaction object.

    Currently implemented GET filter parameters:
        * fullDetails
        * limit
        * offset

    See Mambu official developer documentation for further details, and
    info on parameters that may be implemented here in the future.

    .. todo:: status API V2: NOT compatible . Apparently changing
              'savings' to 'deposits' may work
    """
    getparams = []
    if kwargs:
        getparams = _get_parameters_url(kwargs)

    url = (
        getmambuurl(*args, **kwargs)
        + "savings/"
        + idcred
        + "/transactions/"
        + ("" if len(getparams) == 0 else "?" + "&".join(getparams))
    )
    return url


def getsavingstransactionssearchurl(idcred, *args, **kwargs):
    """Request saving transaction URL.

    idcred is not used - search is rather defined via POST filter parameter.

    Currently implemented GET filter parameters:
        * fullDetails
        * limit
        * offset

    See Mambu official developer documentation for further details, and
    info on parameters that may be implemented here in the future.

    .. todo:: status API V2: NOT compatible . Apparently changing
              'savings' to 'deposits' may work
    """
    getparams = []
    if kwargs:
        getparams = _get_parameters_url(kwargs)

    url = (
        getmambuurl(*args, **kwargs)
        + "savings/transactions/search"
        + ("" if len(getparams) == 0 else "?" + "&".join(getparams))
    )
    return url


def gettransactionchannelsurl(idcred, *args, **kwargs):
    """Transaction channels URL.

    If idcred is set, you'll get a response adequate for a
    MambuTransactionChannel object.

    Currently implemented GET filter parameters:
        * fullDetails
        * limit
        * offset

    See Mambu official developer documentation for further details, and
    info on parameters that may be implemented here in the future.

    .. todo:: status API V2: NOT compatible . Apparently changing
              URL to /organization/transactionchannels may work
    """
    getparams = []
    if kwargs:
        getparams = _get_parameters_url(kwargs)

    url = (
        getmambuurl(*args, **kwargs)
        + "transactionchannels/"
        + idcred
        + ("" if len(getparams) == 0 else "?" + "&".join(getparams))
    )
    return url


def getloansurl(idcred, *args, **kwargs):
    """Request Loans URL.

    If idcred is set, you'll get a response adequate for a MambuLoan object.
    If not set, you'll get a response adequate for a MambuLoans object.
    See mambuloan module and pydoc for further information.

    Currently implemented filter parameters:
    * fullDetails
    * accountState
    * branchId
    * centreId
    * creditOfficerUsername
    * fixedDaysOfMonth
    * startDate
    * endDate
    * limit
    * offset

    See Mambu official developer documentation for further details, and
    info on parameters that may be implemented here in the future.

    .. todo:: status API V2: GET compatible
    """
    getparams = []
    if kwargs:
        getparams = _get_parameters_url(kwargs)

    idcredparam = "" if idcred == "" else "/" + idcred
    url = (
        getmambuurl(*args, **kwargs)
        + "loans"
        + idcredparam
        + ("" if len(getparams) == 0 else "?" + "&".join(getparams))
    )
    return url


def getloanscustominformationurl(idcred, customfield="", *args, **kwargs):
    """Request Loan Custom Information URL.

    See Mambu official developer documentation for further details, and
    info on parameters that may be implemented here in the future.

    .. todo:: status API V2: NOT compatible
    """
    loanidparam = "/" + idcred
    url = (
        getmambuurl(*args, **kwargs)
        + "loans"
        + loanidparam
        + "/custominformation"
        + (("/" + customfield) if customfield else "")
    )
    return url


def getusercustominformationurl(iduser, customfield="", *args, **kwargs):
    """Request User Custom Information URL.

    See Mambu official developer documentation for further details, and
    info on parameters that may be implemented here in the future.

    .. todo:: status API V2: NOT compatible
    """
    useridparam = "/" + iduser
    url = (
        getmambuurl(*args, **kwargs)
        + "users"
        + useridparam
        + "/custominformation"
        + (("/" + customfield) if customfield else "")
    )
    return url


def getgroupurl(idgroup, *args, **kwargs):
    """Request Groups URL.

    If idgroup is set, you'll get a response adequate for a MambuGroup object.
    If not set, you'll get a response adequate for a MambuGroups object.
    See mambugroup module and pydoc for further information.

    Currently implemented filter parameters:
    * fullDetails
    * creditOfficerUsername
    * branchId
    * centreId
    * limit
    * offset

    See Mambu official developer documentation for further details, and
    info on parameters that may be implemented here in the future.

    .. todo:: status API V2: GET compatible

    .. todo:: status API V2: other opers: needs test
    """
    getparams = []
    if kwargs:
        getparams = _get_parameters_url(kwargs)

    groupIdparam = "" if idgroup == "" else "/" + idgroup

    url = (
        getmambuurl(*args, **kwargs)
        + "groups"
        + groupIdparam
        + ("" if len(getparams) == 0 else "?" + "&".join(getparams))
    )
    return url


def getgrouploansurl(idgroup, *args, **kwargs):
    """Request Group loans URL.

    How to use it? By default MambuLoan uses getloansurl as the urlfunc.
    Override that behaviour by sending getgrouploansurl (this function)
    as the urlfunc to the constructor of MambuLoans (note the final 's')
    and voila! you get the Loans just for a certain group.

    If idgroup is set, you'll get a response adequate for a
    MambuLoans object.

    If not set, you'll get a Jar Jar Binks object, or something quite
    strange and useless as JarJar. A MambuError must likely since I
    haven't needed it for anything but for loans of one and just
    one group.

    See mambugroup module and pydoc for further information.

    Currently implemented filter parameters:
    * accountState

    See Mambu official developer documentation for further details, and
    info on parameters that may be implemented here in the future.

    .. todo:: status API V2: NOT compatible
    """
    getparams = []
    if kwargs:
        getparams = _get_parameters_url(kwargs)

    groupIdparam = "/" + idgroup
    url = (
        getmambuurl(*args, **kwargs)
        + "groups"
        + groupIdparam
        + "/loans"
        + ("" if len(getparams) == 0 else "?" + "&".join(getparams))
    )
    return url


def getgroupcustominformationurl(idgroup, customfield="", *args, **kwargs):
    """Request Group Custom Information URL.

    See mambugroup module and pydoc for further information.

    See Mambu official developer documentation for further details, and
    info on parameters that may be implemented here in the future.

    .. todo:: status API V2: NOT compatible
    """
    groupIdparam = "/" + idgroup
    url = (
        getmambuurl(*args, **kwargs)
        + "groups"
        + groupIdparam
        + "/custominformation"
        + (("/" + customfield) if customfield else "")
    )
    return url


def gettransactionsurl(idcred, *args, **kwargs):
    """Request loan Transactions URL.

    If idcred is set, you'll get a response adequate for a
    MambuTransactions object. There's a MambuTransaction object too, but
    you'll get a list first and each element of it will be automatically
    converted to a MambuTransaction object that you may use.

    If not set, you'll get a Jar Jar Binks object, or something quite
    strange and useless as JarJar. A MambuError must likely since I
    haven't needed it for anything but for transactions of one and just
    one loan account.

    See mambutransaction module and pydoc for further information.

    Currently implemented filter parameters:
    * limit
    * offset

    See Mambu official developer documentation for further details, and
    info on parameters that may be implemented here in the future.

    .. todo:: status API V2: NOT compatible.
              Repayments/fees/adjustments/disbursements have their own
              endpoint each
    """
    getparams = []
    if kwargs:
        getparams = _get_parameters_url(kwargs)
    url = (
        getmambuurl(*args, **kwargs)
        + "loans/"
        + idcred
        + "/transactions"
        + ("" if len(getparams) == 0 else "?" + "&".join(getparams))
    )
    return url


def getclienturl(idclient, *args, **kwargs):
    """Request Clients URL.

    If idclient is set, you'll get a response adequate for a MambuClient object.
    If not set, you'll get a response adequate for a MambuClients object.
    See mambuclient module and pydoc for further information.

    Currently implemented filter parameters:
    * fullDetails
    * firstName
    * lastName
    * idDocument
    * birthdate
    * state
    * limit
    * offset

    See Mambu official developer documentation for further details, and
    info on parameters that may be implemented here in the future.

    .. todo:: status API V2: GET compatible

    .. todo:: status API V2: other opers: needs test
    """
    getparams = []
    if kwargs:
        getparams = _get_parameters_url(kwargs)

    clientIdparam = "" if idclient == "" else "/" + idclient
    url = (
        getmambuurl(*args, **kwargs)
        + "clients"
        + clientIdparam
        + ("" if len(getparams) == 0 else "?" + "&".join(getparams))
    )
    return url


def getclientloansurl(idclient, *args, **kwargs):
    """Request Client loans URL.

    How to use it? By default MambuLoan uses getloansurl as the urlfunc.
    Override that behaviour by sending getclientloansurl (this function)
    as the urlfunc to the constructor of MambuLoans (note the final 's')
    and voila! you get the Loans just for a certain client.

    If idclient is set, you'll get a response adequate for a
    MambuLoans object.

    If not set, you'll get a Jar Jar Binks object, or something quite
    strange and useless as JarJar. A MambuError must likely since I
    haven't needed it for anything but for loans of one and just
    one client.

    See mambuloan module and pydoc for further information.

    Currently implemented filter parameters:
    * accountState

    See Mambu official developer documentation for further details, and
    info on parameters that may be implemented here in the future.

    .. todo:: status API V2: NOT compatible
    """
    getparams = []
    if kwargs:
        getparams = _get_parameters_url(kwargs)

    clientIdparam = "/" + idclient
    url = (
        getmambuurl(*args, **kwargs)
        + "clients"
        + clientIdparam
        + "/loans"
        + ("" if len(getparams) == 0 else "?" + "&".join(getparams))
    )
    return url


def getclientcustominformationurl(idclient, customfield="", *args, **kwargs):
    """Request Client Custom Information URL.

    See mambugroup module and pydoc for further information.

    See Mambu official developer documentation for further details, and
    info on parameters that may be implemented here in the future.

    .. todo:: status API V2: NOT compatible
    """
    clientIdparam = "/" + idclient
    url = (
        getmambuurl(*args, **kwargs)
        + "clients"
        + clientIdparam
        + "/custominformation"
        + (("/" + customfield) if customfield else "")
    )
    return url


def getuserurl(iduser, *args, **kwargs):
    """Request Users URL.

    If iduser is set, you'll get a response adequate for a MambuUser object.
    If not set, you'll get a response adequate for a MambuUsers object.
    See mambuuser module and pydoc for further information.

    Currently implemented filter parameters:
    * fullDetails
    * branchId
    * limit
    * offset

    See Mambu official developer documentation for further details, and
    info on parameters that may be implemented here in the future.

    .. todo:: status API V2: compatible
    """
    getparams = []
    if kwargs:
        getparams = _get_parameters_url(kwargs)

    useridparam = "" if iduser == "" else "/" + iduser
    url = (
        getmambuurl(*args, **kwargs)
        + "users"
        + useridparam
        + ("" if len(getparams) == 0 else "?" + "&".join(getparams))
    )
    return url


def getproductsurl(idproduct, *args, **kwargs):
    """Request loan Products URL.

    If idproduct is set, you'll get a response adequate for a MambuProduct object.
    If not set, you'll get a response adequate for a MambuProducts object.
    See mambuproduct module and pydoc for further information.

    No current implemented filter parameters.

    See Mambu official developer documentation for further details, and
    info on parameters that may be implemented here in the future.

    .. todo:: status API V2: NOT compatible. Current (2020-03) v2 only
              supports GETting a SINGLE product, not all of them
    """
    productidparam = "" if idproduct == "" else "/" + idproduct
    url = getmambuurl(*args, **kwargs) + "loanproducts" + productidparam
    return url


def gettasksurl(dummyId="", *args, **kwargs):
    """Request Tasks URL.

    dummyId is used because MambuStruct always requires an Id from an
    entity, but the Mambu API doesn't requires it for Tasks, because of
    that dummyId defaults to '', but in practice it is never used (if
    someone sends dummyId='someId' nothing happens). The fact of forcing
    to send an entid is a technical debt that should be payed.

    Currently implemented filter parameters:
    * username
    * clientId
    * groupId
    * status
    * limit
    * offset

    Mambu REST API defaults to open when status not provided. Here we
    are just making that explicit always defaulting status to 'OPEN'

    See Mambu official developer documentation for further details, and
    info on parameters that may be implemented here in the future.

    .. todo:: status API V2: GET compatible
    """
    getparams = []
    if kwargs:
        try:
            getparams.append("status=%s" % kwargs["status"])
            del kwargs["status"]
        except Exception:
            getparams.append("status=OPEN")
        getparams.extend(_get_parameters_url(kwargs))

    url = (
        getmambuurl(*args, **kwargs)
        + "tasks"
        + ("" if len(getparams) == 0 else "?" + "&".join(getparams))
    )
    return url


def getactivitiesurl(dummyId="", *args, **kwargs):
    """Request Activities URL.

    dummyId is used because MambuStruct always requires an Id from an
    entity, but the Mambu API doesn't requires it for Activities,
    because of that dummyId defaults to '', but in practice it is never
    used (if someone sends dummyId='someId' nothing happens). The fact
    of forcing to send an entid is a technical debt that should be
    payed.

    Currently implemented filter parameters:
    * from
    * to
    * branchID
    * clientId
    * centreID
    * userID
    * loanAccountId
    * groupId
    * limit

    See Mambu official developer documentation for further details, and
    info on parameters that may be implemented here in the future.

    .. todo:: status API V2: NOT compatible. Apparently not yet implemented
    """
    from datetime import datetime

    getparams = []
    if kwargs:
        try:
            getparams.append("from=%s" % kwargs["fromDate"])
            del kwargs["fromDate"]
        except Exception:
            getparams.append("from=%s" % "1900-01-01")

        try:
            getparams.append("to=%s" % kwargs["toDate"])
            del kwargs["toDate"]
        except Exception:
            hoy = datetime.now().strftime("%Y-%m-%d")
            getparams.append("to=%s" % hoy)

        getparams.extend(_get_parameters_url(kwargs))

    url = (
        getmambuurl(*args, **kwargs)
        + "activities"
        + ("" if len(getparams) == 0 else "?" + "&".join(getparams))
    )
    return url


def getrolesurl(idrole="", *args, **kwargs):
    """Request Roles URL.

    If idrole is set, you'll get a response adequate for a MambuRole
    object.  If not set, you'll get a response adequate for a MambuRoles
    object too.  See mamburoles module and pydoc for further
    information.

    See Mambu official developer documentation for further details.

    .. todo:: status API V2: NOT compatible. Apparently not yet implemented
    """
    url = getmambuurl(*args, **kwargs) + "userroles" + (("/" + idrole) if idrole else "")
    return url


def getpostdocumentsurl(identity="", *args, **kwargs):
    """Request to post documents URL.

    See Mambu official developer documentation for further details.
    https://support.mambu.com/docs/attachments-api

    """
    url = getmambuurl(*args, **kwargs) + "documents"
    return url



"""Schema table for Mambu Branches.

.. autosummary::
   :nosignatures:
   :toctree: _autosummary

.. todo:: this is just a basic schema for centres. Some fields
          are missing.
"""

from sqlalchemy import Column, DateTime, ForeignKey, String
from sqlalchemy.orm import relationship

from . import schema_orm as orm
from .schema_branches import Branch

dbname = orm.dbname
session = orm.session
Base = orm.Base


class Centre(Base):
    """Centre table.
    Related with client
    """

    __tablename__ = "centre"
    __table_args__ = {"schema": dbname, "keep_existing": True}

    # Columns
    encodedKey = Column(String, primary_key=True)
    id = Column(String, index=True, unique=True)
    name = Column(String)
    notes = Column(String)
    state = Column(String)
    creationDate = Column(DateTime)
    lastModifiedDate = Column(DateTime)

    # Relationships
    assignedBranchKey = Column(String, ForeignKey(Branch.encodedKey))
    loans = relationship("LoanAccount", back_populates="centre")
    branch = relationship(Branch, back_populates="centres")
    groups = relationship("Group", back_populates="centre")

    def __repr__(self):
        return "<Centre(id={}, name={})>".format(self.id, self.name)



"""`SQLAlchemy <http://www.sqlalchemy.org/>`_ ORM for a Mambu Database backup.

.. autosummary::
   :nosignatures:
   :toctree: _autosummary

This is the basic module that holds a default session (and a default
Session maker and a connected engine) which is used on every ORM
module on MambuPy.

The default engine is connected to a database according to what you
configured on mambuconfig module.

This module also holds the Base declarative_base used for every ORM on
MambuPy.

Every ORM module on MambuPy (named schema_*.py) should ideally get the
default session, Base and etc. from here, allowing the user to use the
same session and etc. for the day-to-day work. If you chose to use
another session, please mind that you should replace the session on
every schema_*.py you need.

This last requirement also applies for the Base, or for the engine and
the sessionmaker for that matter.
"""
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker, scoped_session

from sqlalchemy import create_engine
from sqlalchemy.pool import NullPool

from ..mambuconfig import dbeng, dbhost, dbname, dbport, dbpwd, dbuser


# Connects to DB
def connect_db(
    engine=dbeng,
    user=dbuser,
    password=dbpwd,
    host=dbhost,
    port=dbport,
    database=dbname,
    params="?charset=utf8&use_unicode=1",
    echoopt=False,
):
    """Connect to database utility function.

    Uses SQLAlchemy ORM library.

    Useful when using schema modules in MambuPy
    """
    return create_engine(
        "%s://%s:%s@%s:%s/%s%s" % (engine, user, password, host, port, database, params),
        poolclass=NullPool,
        isolation_level="READ UNCOMMITTED",
        echo=echoopt,
    )


engine = connect_db()
"""Database engine, connecting with default parameters by default.
"""

session_factory = sessionmaker(bind=engine)
"""Sessionmaker object, create factory sessions.
"""

Session = scoped_session(session_factory)
"""Scoped session object.
"""

session = Session()
"""Default session created here.
"""

Base = declarative_base()
"""Declarative base for models.
"""



"""*MambuPy*, an API library to access `Mambu <https://www.mambu.com/>`_ objects.

.. autosummary::
   :toctree: _autosummary

    MambuPy.mambuconfig
    MambuPy.mambuutil
    MambuPy.api
    MambuPy.api.connector
    MambuPy.orm


Currently, there are two different ways to access Mambu objects:

1) Objects using `Mambu REST API
   <https://developer.mambu.com/customer/en/portal/articles/1162276-rest-apis-overview/>`_
   , they live at the :any:`MambuPy.api` package

2) An ORM using a `DB backup retrieved from Mambu
   <https://support.mambu.com/docs/mambu-data-dictionary>`_
   , they live at the :any:`MambuPy.orm` package

.. note::

   The `Mambu Developers Center site <https://developer.mambu.com/>`_ holds the
   current documentation for the latest version of Mambu. *MambuPy* will try to
   keep up with what Mambu updates according to such information.

TODOS
=====

.. todo:: Unit testing of some modules is currently very basic. The purpose is
          to achive TDD when implementing features or correcting bugs.
"""

import sys
from importlib.abc import MetaPathFinder
from importlib.util import spec_from_file_location

__version__ = "2.2.1"
"""The version of this module."""


class CaseInsensitiveFinder(MetaPathFinder):
    def find_spec(self, fullname, path, target=None):
        if fullname.lower() == "mambupy":
            # Get the actual module file location
            return spec_from_file_location(fullname, __file__)
        return None


# Register the finder
sys.meta_path.insert(0, CaseInsensitiveFinder())



"""Schema table for Mambu Branches.

.. autosummary::
   :nosignatures:
   :toctree: _autosummary

.. todo:: this is just a basic schema for branches. Some fields
          are missing.
"""

from sqlalchemy import Column, DateTime, ForeignKey, Integer, Numeric, String
from sqlalchemy.orm import relationship

from . import schema_orm as orm

dbname = orm.dbname
session = orm.session
Base = orm.Base


class Branch(Base):
    """Branch table.
    Related with client
    """

    __tablename__ = "branch"
    __table_args__ = {"schema": dbname, "keep_existing": True}

    # Columns
    encodedKey = Column(String, primary_key=True)
    id = Column(String, index=True, unique=True)
    name = Column(String)
    notes = Column(String)
    phoneNumber = Column(String)
    emailAddress = Column(String)
    state = Column(String)
    creationDate = Column(DateTime)
    lastModifiedDate = Column(DateTime)

    # Relationships
    addresses = relationship(
        "Address",
        back_populates="branch",
        foreign_keys="Address.parentKey",
        primaryjoin="Address.parentKey == Branch.encodedKey",
    )
    customInformation = relationship(
        "CustomFieldValue",
        back_populates="branch",
        foreign_keys="CustomFieldValue.parentKey",
        primaryjoin="CustomFieldValue.parentKey == Branch.encodedKey",
    )
    loans = relationship("LoanAccount", back_populates="branch")
    activities = relationship("Activity", back_populates="branch")
    clients = relationship("Client", back_populates="branch")
    groups = relationship("Group", back_populates="branch")
    users = relationship("User", back_populates="branch")
    centres = relationship("Centre", back_populates="branch")

    def __repr__(self):
        return "<Branch(id={}, name={})>".format(self.id, self.name)



""" MambuPy's ORM package.

A `SQLAlchemy <http://www.sqlalchemy.org/>`_ ORM using a `DB backup retrieved from Mambu
<https://support.mambu.com/docs/mambu-data-dictionary>`_

.. autosummary::
   :toctree: _autosummary

    MambuPy.orm.schema_orm
    MambuPy.orm.schema_mambu
    MambuPy.orm.schema_clients
    MambuPy.orm.schema_groups
    MambuPy.orm.schema_loans
    MambuPy.orm.schema_users
    MambuPy.orm.schema_branches
    MambuPy.orm.schema_addresses
    MambuPy.orm.schema_activities
    MambuPy.orm.schema_customfields
    MambuPy.orm.schema_dummies


Lives under the :py:mod:`MambuPy.orm` package

.. todo:: Implement a lot of lacking fields on the currently available tables.
.. todo:: Implement a lot of lacking tables from the Mambu DB schema.
"""

from .. import mambuconfig, mambuutil



"""Schema tables for Mambu Activities.

.. autosummary::
   :nosignatures:
   :toctree: _autosummary
"""

from sqlalchemy import (Column, DateTime, ForeignKey, Integer, Numeric, String,
                        Table)
from sqlalchemy.orm import backref, relationship

from . import schema_orm as orm
from .schema_branches import Branch
from .schema_clients import Client
from .schema_groups import Group
from .schema_loans import LoanAccount
from .schema_users import User

dbname = orm.dbname
session = orm.session
Base = orm.Base


class Activity(Base):
    """Activity table."""

    __tablename__ = "activity"
    __table_args__ = {"schema": dbname, "keep_existing": True}

    # Columns
    encodedKey = Column(String, primary_key=True)
    lineOfCreditKey = Column(String)
    centreKey = Column(String)
    loanProductKey = Column(String)
    savingsAccountKey = Column(String)
    savingsProductKey = Column(String)
    assignedCentreKey = Column(String)
    taskKey = Column(String)
    glAccountKey = Column(String)
    glAccountsClosureKey = Column(String)
    entityKey = Column(String)
    transactionId = Column(Integer)
    entityType = Column(String)
    parent_key = Column(String)
    fieldChangeName = Column(String)
    activityChanges_integer_idx = Column(Integer)

    timestamp = Column(DateTime)
    notes = Column(String)
    type = Column(String)

    # Relationships
    loanAccountKey = Column(String, ForeignKey(LoanAccount.encodedKey))
    branchKey = Column(String, ForeignKey(Branch.encodedKey))
    clientKey = Column(String, ForeignKey(Client.encodedKey))
    groupKey = Column(String, ForeignKey(Group.encodedKey))
    userKey = Column(String, ForeignKey(User.encodedKey))
    assignedUserKey = Column(String, ForeignKey(User.encodedKey))

    loan = relationship(LoanAccount, back_populates="activities")
    branch = relationship(Branch, back_populates="activities")
    client = relationship(Client, back_populates="activities")
    group = relationship(Group, back_populates="activities")
    user = relationship(
        User,
        primaryjoin="User.encodedKey==Activity.userKey",
        back_populates="activities",
    )
    assignedUser = relationship(
        User,
        primaryjoin="User.encodedKey==Activity.assignedUserKey",
        back_populates="assignedActivities",
    )
    fieldChanges = relationship("FieldChangeItem", back_populates="activity")

    def __repr__(self):
        return "<Activity(type=%s, timestamp=%s)>" % (
            self.type,
            self.timestamp.strftime("%Y%m%d"),
        )


class FieldChangeItem(Base):
    """Fieldchangeitem table."""

    __tablename__ = "fieldchangeitem"
    __table_args__ = {"schema": dbname, "keep_existing": True}

    # Columns
    id = Column(String, primary_key=True)
    fieldChangeName = Column(String)
    fieldDetailName = Column(String)
    fieldDetailKey = Column(String)
    newValue = Column(String)
    originalValue = Column(String)
    fieldchanges_integer_idx = Column(Integer)

    # Relationships
    fieldchanges_encodedkey_own = Column(String, ForeignKey(Activity.encodedKey))
    activity = relationship(Activity, back_populates="fieldChanges")

    def __repr__(self):
        return (
            "<FieldChangeItem(newValue=%s, originalValue=%s, fieldChangeName=%s)>"
            % (self.originalValue, self.newValue, self.fieldChangeName)
        )



""" MambuPy's Mambu REST API v2 package.

Objects using `Mambu REST API <https://api.mambu.com/?python>`_

.. autosummary::
   :toctree: _autosummary

    MambuPy.api.classes
    MambuPy.api.entities
    MambuPy.api.interfaces
    MambuPy.api.vos
    MambuPy.api.mambustruct
    MambuPy.api.mambuproduct
    MambuPy.api.mambuloan
    MambuPy.api.mambutransaction
    MambuPy.api.mambugroup
    MambuPy.api.mambuclient
    MambuPy.api.mambubranch
    MambuPy.api.mambucentre
    MambuPy.api.mambuuser
    MambuPy.api.mambutask
    MambuPy.api.mambucustomfield

Entities use the Mixin pattern to acquire functionality.
"""



"""MambuClient entity: a MambuEntity struct for Clients.

.. autosummary::
   :nosignatures:
   :toctree: _autosummary
"""

from .entities import (
    MambuEntity,
    MambuEntityWritable,
    MambuEntityAttachable,
    MambuEntitySearchable,
    MambuEntityCommentable,
)
from .interfaces import MambuHolder


class MambuClient(
    MambuEntity,
    MambuEntityWritable,
    MambuEntityAttachable,
    MambuEntitySearchable,
    MambuEntityCommentable,
    MambuHolder,
):
    """MambuClient entity"""

    _prefix = "clients"
    """prefix constant for connections to Mambu"""

    _default_tzattrs = {
        "birthDate": None,
        "groupKeys": [None],
        "addresses": [{}],
        "idDocuments": [{}],
    }

    _filter_keys = [
        "branchId",
        "centreId",
        "creditOfficerUsername",
        "firstName",
        "lastName",
        "idNumber",
        "state",
        "birthDate",
    ]
    """allowed filters for get_all filtering"""

    _sortBy_fields = [
        "creationDate",
        "lastModifiedDate",
        "firstName",
        "lastName",
    ]
    """allowed fields for get_all sorting"""

    _ownerType = "CLIENT"
    """owner type of this entity"""

    _vos = [("addresses", "MambuAddress"), ("idDocuments", "MambuIDDocument")]
    """2-tuples of elements and Value Objects"""

    _entities = [
        ("groupKeys", "mambugroup.MambuGroup", "groups"),
        ("assignedUserKey", "mambuuser.MambuUser", "assignedUser"),
        ("assignedBranchKey", "mambubranch.MambuBranch", "assignedBranch"),
        ("assignedCentreKey", "mambucentre.MambuCentre", "assignedCentre"),
    ]
    """3-tuples of elements and Mambu Entities"""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._attachments = {}

    def _delete_for_creation(self):
        """Deletes extra fields from Mambu unusable for entity
        creation."""
        try:
            del self._attrs["approvedDate"]
        except KeyError:
            pass



"""Connectors to Mambu.

Currently supports REST.

.. autosummary::
   :nosignatures:
   :toctree: _autosummary
"""

import base64
import copy
import json
import mimetypes
import os
import re
import uuid

import requests
from requests_toolbelt.multipart.encoder import MultipartEncoder
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry

from .mambuconnector import MambuConnector, MambuConnectorReader, MambuConnectorWriter
from MambuPy.mambuutil import (
    ALLOWED_UPLOAD_MIMETYPES,
    DETAILSLEVEL,
    MAX_UPLOAD_SIZE,
    PAGINATIONDETAILS,
    SEARCH_OPERATORS,
    OUT_OF_BOUNDS_PAGINATION_LIMIT_VALUE,
    UPLOAD_FILENAME_INVALID_CHARS,
    MambuCommError,
    MambuError,
    MambuPyError,
    apipwd,
    apiurl,
    apiuser,
    setup_logging,
    activate_request_session_objects,
)


logger = setup_logging(__name__)


def _configure_retry_strategy(session, retries=5):
    """Configure retry strategy for a session.

    Args:
        session (requests.Session): The session to configure
        retries (int, optional): Number of retries. Defaults to 5.
    """
    retry_strategy = Retry(
        total=retries,
        status_forcelist=[429, 500, 502, 503, 504],
        backoff_factor=1,
        allowed_methods=[
            "HEAD",
            "GET",
            "OPTIONS",
            "POST",
            "PUT",
            "DELETE",
            "TRACE",
            "PATCH",
        ],
    )
    adapter = HTTPAdapter(max_retries=retry_strategy)
    session.mount("https://", adapter)
    session.mount("http://", adapter)


class SessionSingleton:
    """Singleton class to manage HTTP sessions for MambuPy.

    This class ensures that only one requests.Session object is created and reused
    across all requests to the Mambu API. This helps improve performance by reusing
    TCP connections.

    Example:
        >>> session = SessionSingleton()
        >>> session.get_session()  # Returns a requests.Session object
        >>> session2 = SessionSingleton()
        >>> session2.get_session()  # Returns the same requests.Session object
    """

    __instance = None
    __session = None
    _RETRIES = 5

    def __new__(cls):
        """Create a new instance of SessionSingleton if one doesn't exist.

        Returns:
            SessionSingleton: The singleton instance.
        """
        if cls.__instance is None:
            cls.__instance = super(SessionSingleton, cls).__new__(cls)
        return cls.__instance

    def get_session(self):
        """Get the requests.Session object.

        If no session exists, create a new one. Otherwise, return the existing one.

        Returns:
            requests.Session: The session object to use for HTTP requests.
        """
        if self.__session is None:
            self.__session = requests.Session()
            _configure_retry_strategy(self.__session, self._RETRIES)
        return self.__session


class MambuConnectorREST(MambuConnector, MambuConnectorReader, MambuConnectorWriter):
    """A connector for Mambu REST API"""

    _RETRIES = 5
    _tenant = ""
    _headers = {}

    def __init__(self, user=apiuser, pwd=apipwd, url=apiurl, **kwargs):
        self._headers = {
            "Accept": "application/vnd.mambu.v2+json",
        }
        self.__set_authorization_header(user, pwd)
        self.__set_url(url)
        self._session = None
        if activate_request_session_objects.lower() == "true":
            self._session = SessionSingleton().get_session()

    def __set_authorization_header(self, user, pwd):
        self._headers["Authorization"] = "Basic {}".format(
            base64.b64encode(bytes("{}:{}".format(user, pwd), "utf-8")).decode()
        )

    def __set_url(self, url):
        self._tenant = url

    def __request_headers(self, method, content_type):
        headers = copy.copy(self._headers)

        if method in ["POST", "PATCH", "PUT"]:
            headers["Content-Type"] = "application/json"
            headers["Idempotency-Key"] = str(uuid.uuid4())
        if content_type:
            headers["Content-Type"] = content_type

        return headers

    def __request_params(self, params):
        if not params:
            params = {}
        return params

    def __request_data(self, data):
        if data is not None:
            try:
                data = json.dumps(data)
            except TypeError:
                data = data
        return data

    def __request(self, method, url, params=None, data=None, content_type=None):
        """requests an url.

        Args:
          method (str): HTTP method for the request
          url (str): URL for the request
          params (dict): query parameters
          data (serializable list of dicts or dict alone): request body
          content_type (str): an alternative Content-Type to send in headers

        Returns:
          response content (json)

        Raises:
          `MambuError`: in case of 400 or 500 response codes
        """
        headers = self.__request_headers(method, content_type)
        params = self.__request_params(params)
        data = self.__request_data(data)

        resp = ""
        try:
            logger.debug(
                "about to make %s request: \
url %s, params %s, data %s, headers %s",
                method,
                url,
                [(k, v) for k, v in params.items() if k not in ["pwd"]],
                data,
                [(k, v) for k, v in headers.items() if k != "Authorization"],
            )
            if self._session:
                resp = self._session.request(method, url, params=params, data=data, headers=headers)
            else:
                http = requests.Session()
                _configure_retry_strategy(http, self._RETRIES)
                resp = http.request(method, url, params=params, data=data, headers=headers)
            resp.raise_for_status()
        except requests.exceptions.HTTPError as httperr:
            logger.warning(
                "%s on %s request: params %s, data %s, headers %s",
                str(httperr),
                method,
                params,
                data,
                [(k, v) for k, v in headers.items() if k != "Authorization"],
            )
            if hasattr(resp, "content"):  # pragma: no cover
                logger.warning("HTTPError, resp content: %s", resp.content)
            try:
                content = json.loads(resp.content.decode())
            except ValueError:
                # in case resp.content doesn't conforms to json
                content = {
                    "errors": [
                        {
                            "errorCode": "UNKNOWN",
                            "errorReason": resp.content.decode(),
                        },
                    ]
                }
            try:
                error = content["errors"][0]
            except KeyError:
                error = content
            raise MambuError(
                "{} ({}) - {}{}".format(
                    error["errorCode"] if "errorCode" in error else error["returnCode"],
                    resp.status_code,
                    (
                        error["errorReason"]
                        if "errorReason" in error
                        else error["returnStatus"]
                    ),
                    " (" + error["errorSource"] + ")" if "errorSource" in error else "",
                )
            )
        except requests.exceptions.RetryError as rerr:  # pragma: no cover
            logger.error(
                "%s MambuCommError on %s request: url %s, params %s, data %s, headers %s",
                str(rerr),
                method,
                url,
                params,
                data,
                [(k, v) for k, v in headers.items() if k != "Authorization"],
            )
            raise MambuCommError(
                "COMM Error: I cannot communicate with Mambu: {}".format(rerr)
            )
        except Exception as ex:
            logger.exception(
                "%s Exception (%s) on %s request: url %s, params %s, data %s, headers %s",
                str(ex),
                resp,
                method,
                url,
                params,
                data,
                [(k, v) for k, v in headers.items() if k != "Authorization"],
            )
            if hasattr(resp, "content"):  # pragma: no cover
                logger.error("Exception, resp content: %s", resp.content)
            raise MambuCommError("Unknown comm error with Mambu: {}".format(ex))

        logger.debug("response %s to %s:\n%s", resp, method, resp.content)

        return resp.content

    def __list_request_args(self, params):
        if not params:
            params = {}
        if "limit" in params and params["limit"] is not None:
            ini_limit = params["limit"]
        else:
            ini_limit = 0
        if "offset" in params and params["offset"] is not None:
            offset = params["offset"]
        else:
            offset = 0

        return (params, ini_limit, offset)

    def __list_request_cat_response(self, list_resp, resp):
        if list_resp == b"":
            list_resp = resp
        elif len(json.loads(resp.decode())) > 0:
            list_resp = list_resp[:-1] + b"," + resp[1:]

        return list_resp

    def __list_request(self, method, url, params=None, data=None):
        """Request for a list, appending responses with limit and offset.

        Makes several requests adjusting limits and offsets, appending
        responses, just as if you have made a single request with a
        big response.

        Useful for services where you have a Maximum limit of response
        elements but wish to make a single call as if doing a single
        request (but not, you make as many as necessary until covering
        the given limit).

        Args:
          method (str): HTTP method for the request
          url (str): URL for the request
          params (dict): query parameters
          data (serializable list of dicts or dict alone): request body
        """
        (params, ini_limit, offset) = self.__list_request_args(params)

        window = True
        list_resp = b""
        while window:
            if not ini_limit or ini_limit > OUT_OF_BOUNDS_PAGINATION_LIMIT_VALUE:
                limit = OUT_OF_BOUNDS_PAGINATION_LIMIT_VALUE
            else:
                limit = ini_limit

            params["offset"] = offset
            params["limit"] = limit if limit != 0 else None

            resp = self.__request(
                method, url, params=copy.copy(params), data=copy.copy(data)
            )

            jsonresp = list(json.loads(resp.decode()))
            if len(jsonresp) < limit:
                window = False

            list_resp = self.__list_request_cat_response(list_resp, resp)

            # next window, moving offset...
            offset = offset + limit
            if ini_limit:
                ini_limit -= limit
                if ini_limit <= 0:
                    window = False

        return list_resp

    def __validate_query_params(self, **kwargs):
        """Validate query params

        Args:
          **kwargs (dict): query params

        Returns:
          params (dict): validated params for a query
        """
        params = {}

        if "offset" in kwargs and kwargs["offset"] is not None:
            if not isinstance(kwargs["offset"], int):
                raise MambuPyError("offset must be integer")
            params["offset"] = kwargs["offset"]

        if "limit" in kwargs and kwargs["limit"] is not None:
            if not isinstance(kwargs["limit"], int):
                raise MambuPyError("limit must be integer")
            params["limit"] = kwargs["limit"]

        if "paginationDetails" in kwargs:
            if kwargs["paginationDetails"] not in PAGINATIONDETAILS:
                raise MambuPyError(
                    "paginationDetails must be in {}".format(PAGINATIONDETAILS)
                )
            params["paginationDetails"] = kwargs["paginationDetails"]

        if "detailsLevel" in kwargs:
            if kwargs["detailsLevel"] not in DETAILSLEVEL:
                raise MambuPyError("detailsLevel must be in {}".format(DETAILSLEVEL))
            params["detailsLevel"] = kwargs["detailsLevel"]

        return params

    def __validate_filter_criteria(self, filterCriteria):
        """Validate search filter criteria

        Args:
          filterCriteria (list of dicts): filter criteria

        Returns:
          validated_filterCriteria (list of dicts): validated filter criteria

        Raises:
          MambuPyError (obj): when invalid filterCriteria
        """
        validated_filterCriteria = []
        if not isinstance(filterCriteria, list) or any(
            [criteria for criteria in filterCriteria if not isinstance(criteria, dict)]
        ):
            raise MambuPyError("filterCriteria must be a list of dictionaries")
        for criteria in filterCriteria:
            if (
                "field" not in criteria
                or "operator" not in criteria
                or criteria["operator"] not in SEARCH_OPERATORS
            ):
                raise MambuPyError(
                    "a filterCriteria must have a field and an operator, member of {}".format(
                        SEARCH_OPERATORS
                    )
                )
            validated_filterCriteria.append(criteria)

        return validated_filterCriteria

    def __validate_sorting_criteria(self, sortingCriteria):
        """Validate search sorting criteria

        Args:
          sortingCriteria (dict): sorting criteria

        Returns:
          sortingCriteria (dict): validated sorting criteria

        Raises:
          MambuPyError (obj): when invalid sortingCriteria
        """
        if not isinstance(sortingCriteria, dict):
            raise MambuPyError("sortingCriteria must be a dictionary")
        if (
            "field" not in sortingCriteria
            or "order" not in sortingCriteria
            or sortingCriteria["order"] not in ["ASC", "DESC"]
        ):
            raise MambuPyError(
                "sortingCriteria must have a field and an order, member of {}".format(
                    ["ASC", "DESC"]
                )
            )

        return sortingCriteria

    def mambu_get(self, entid, prefix, detailsLevel="BASIC"):
        """get, a single entity, identified by its entid.

        Args:
          entid (str): ID for the entity
          prefix (str): entity's URL prefix
          detailsLevel (str BASIC/FULL): ask for extra details or not

        Returns:
          response content (str json {})
        """
        params = self.__validate_query_params(detailsLevel=detailsLevel)

        url = "https://{}/api/{}/{}".format(self._tenant, prefix, entid)

        return self.__request("GET", url, params=params)

    def mambu_get_all(
        self,
        prefix,
        filters=None,
        offset=None,
        limit=None,
        paginationDetails="OFF",
        detailsLevel="BASIC",
        sortBy=None,
        **kwargs
    ):
        """get_all, several entities, filtering allowed

        Args:
          prefix (str): entity's URL prefix
          filters (dict): key-value filters (depends on each entity)
          offset (int): pagination, index to start searching
          limit (int): pagination, number of elements to retrieve
          paginationDetails (str ON/OFF): ask for details on pagination
          detailsLevel (str BASIC/FULL): ask for extra details or not
          sortBy (str): ``field1:ASC,field2:DESC``, sorting criteria for
                        results
          kwargs (dict): extra parameters that a specific entity may receive in
                         its get_all method

        Returns:
          response content (str json [])
        """
        params = self.__validate_query_params(
            offset=offset,
            limit=limit,
            paginationDetails=paginationDetails,
            detailsLevel=detailsLevel,
        )
        if kwargs:
            params.update(kwargs)

        if sortBy:
            if not isinstance(sortBy, str) or not re.search(
                r"^(\w+:(ASC|DESC),)*(\w+:(ASC|DESC))$", sortBy
            ):
                raise MambuPyError(
                    "sortBy must be a string with format 'field1:ASC,field2:DESC'"
                )
            params["sortBy"] = sortBy

        if filters:
            if not isinstance(filters, dict):
                raise MambuPyError("filters must be a dictionary")
            params.update(filters)

        url = "https://{}/api/{}".format(self._tenant, prefix)

        return self.__list_request("GET", url, params=params)

    def mambu_search(
        self,
        prefix,
        filterCriteria=None,
        sortingCriteria=None,
        offset=None,
        limit=None,
        paginationDetails="OFF",
        detailsLevel="BASIC",
    ):
        """search, several entities, filtering criteria allowed

        Args:
          prefix (str): entity's URL prefix
          filterCriteria (list of dicts): fields according to
                              LoanAccountFilterCriteria schema
          sortingCriteria (dict): fields according to
                            LoanAccountSortingCriteria
          offset (int): pagination, index to start searching
          limit (int): pagination, number of elements to retrieve
          paginationDetails (str ON/OFF): ask for details on pagination
          detailsLevel (str BASIC/FULL): ask for extra details or not

        Returns:
          response content (str json [])
        """
        params = self.__validate_query_params(
            offset=offset,
            limit=limit,
            paginationDetails=paginationDetails,
            detailsLevel=detailsLevel,
        )

        data = {}

        if filterCriteria is not None:
            data["filterCriteria"] = self.__validate_filter_criteria(filterCriteria)

        if sortingCriteria is not None:
            data["sortingCriteria"] = self.__validate_sorting_criteria(sortingCriteria)

        url = "https://{}/api/{}:search".format(self._tenant, prefix)

        return self.__list_request("POST", url, params=params, data=data)

    def mambu_update(self, entid, prefix, attrs, **kwargs):
        """updates a mambu entity

        Args:
          entid (str): the id or encoded key of the entity owning the document
          prefix (str): entity's URL prefix
          attrs (dict): entity to be updated, complying with Mambu's schemas
          kwargs (dict): keyword arguments for this method.
                         May include a user, pwd and url to connect to Mambu.

        Returns:
          response content (str json {})
        """
        if "url" in kwargs:
            self.__set_url(kwargs["url"])
        if "user" in kwargs and "pwd" in kwargs:
            self.__set_authorization_header(kwargs["user"], kwargs["pwd"])
        url = "https://{}/api/{}/{}".format(self._tenant, prefix, entid)

        return self.__request("PUT", url, data=attrs)

    def mambu_create(self, prefix, attrs, **kwargs):
        """creates a mambu entity

        Args:
          prefix (str): entity's URL prefix
          attrs (dict): entity to be created, complying with Mambu's schemas
          kwargs (dict): keyword arguments for this method.
                         May include a user, pwd and url to connect to Mambu.

        Returns:
          response content (str json {})
        """
        if "url" in kwargs:
            self.__set_url(kwargs["url"])
        if "user" in kwargs and "pwd" in kwargs:
            self.__set_authorization_header(kwargs["user"], kwargs["pwd"])
        url = "https://{}/api/{}".format(self._tenant, prefix)

        return self.__request("POST", url, data=attrs)

    def mambu_patch(self, entid, prefix, fields_ops=None, **kwargs):
        """patches certain parts of a mambu entity

        https://api.mambu.com/?python#tocspatchoperation

        Supported: add, remove, replace (move not yet supported)

        Args:
          entid (str): the id or encoded key of the entity
          prefix (str): entity's URL prefix
          fields_ops (list of tuples): each tuple has:
            OP (str): operation ("ADD", "REPLACE", "REMOVE")
            PATH (str): json pointer referencing the location in the target entity
            VALUE (obj, opc): the value of the field (not for REMOVE op)
          kwargs (dict): keyword arguments for this method.
                         May include a user, pwd and url to connect to Mambu.
        """
        if not fields_ops:
            fields_ops = []

        if "url" in kwargs:
            self.__set_url(kwargs["url"])
        if "user" in kwargs and "pwd" in kwargs:
            self.__set_authorization_header(kwargs["user"], kwargs["pwd"])
        url = "https://{}/api/{}/{}".format(self._tenant, prefix, entid)

        # build data from fields_ops param
        patch_data = []
        for field in fields_ops:
            patch_item = {"op": field[0].strip().lower(), "path": field[1].strip()}
            if field[0] != "REMOVE":
                patch_item["value"] = field[2]
            patch_data.append(patch_item)

        if patch_data:
            return self.__request("PATCH", url, data=patch_data)

    def mambu_delete(self, entid, prefix, **kwargs):
        """deletes a mambu entity

           kwargs (dict): keyword arguments for this method.
                  May include a user, pwd and url to connect to Mambu.
        """
        if "url" in kwargs:
            self.__set_url(kwargs["url"])
        if "user" in kwargs and "pwd" in kwargs:
            self.__set_authorization_header(kwargs["user"], kwargs["pwd"])
        url = "https://{}/api/{}/{}".format(self._tenant, prefix, entid)

        return self.__request("DELETE", url)

    def mambu_upload_document(self, owner_type, entid, filename, name, notes):
        """uploads an attachment to this entity

        Args:
          owner_type (str): the type of the owner of the document
          entid (str): the id or encoded key of the entity owning the document
          filename (str): path and filename of file to upload as attachment
          name (str): name to assign to the attached file in Mambu
          notes (str): notes to associate to the attached file in Mambu

        Returns:
          response content (str json {}) metadata of the attached document
        """
        for char in UPLOAD_FILENAME_INVALID_CHARS:
            if char in os.path.basename(filename):
                raise MambuError("{} name not allowed".format(filename))
        if os.path.basename(filename).count(".") != 1:
            raise MambuError("{} invalid name".format(filename))
        if (
            not mimetypes.guess_type(filename)[0]
            or mimetypes.guess_type(filename)[0].split("/")[1].upper()
            not in ALLOWED_UPLOAD_MIMETYPES
        ):
            raise MambuError("{} mimetype not supported".format(filename))
        if os.path.getsize(filename) > MAX_UPLOAD_SIZE:
            raise MambuError("{} exceeds {} bytes".format(filename, MAX_UPLOAD_SIZE))
        data = {
            "ownerType": owner_type,
            "id": entid,
            "name": name,
            "notes": notes,
            "file": (
                os.path.basename(filename),
                open(filename, "rb"),
                mimetypes.guess_type(filename)[0],
            ),
        }

        encoder = MultipartEncoder(fields=data)

        url = "https://{}/api/documents".format(self._tenant)

        return self.__request(
            "POST", url, data=encoder, content_type=encoder.content_type
        )

    def mambu_get_documents_metadata(
        self,
        entid,
        owner_type,
        offset=None,
        limit=None,
        paginationDetails="OFF",
    ):
        """Gets metadata for all the documents attached to an entity

        Args:
          entid (str): the id or encoded key of the entity owning the document
          owner_type (str): the type of the owner of the document
          offset (int): pagination, index to start searching
          limit (int): pagination, number of elements to retrieve
          paginationDetails (str ON/OFF): ask for details on pagination
        """
        params = self.__validate_query_params(
            offset=offset,
            limit=limit,
            paginationDetails=paginationDetails,
        )

        params.update({"entity": owner_type})
        params.update({"ownerKey": entid})

        url = "https://{}/api/documents/documentsMetadata".format(self._tenant)

        return self.__request("GET", url, params=params)

    def mambu_delete_document(self, documentId):
        """deletes an attachment by its documentId

        Args:
          documentId (str): the id or encodedkey of the document to be deleted
        """
        url = "https://{}/api/documents/{}".format(self._tenant, documentId)

        self.__request("DELETE", url)

    def mambu_loanaccount_getSchedule(self, loanid):
        """Retrieves the installments schedule of a loan account

        Args:
          loanid (str): the id or encoded key of the loan account
        """
        url = "https://{}/api/{}/{}/schedule".format(self._tenant, "loans", loanid)

        return self.__request("GET", url)

    def mambu_loanaccount_writeoff(self, loanid, notes):
        """Writesoff a loan account

        Args:
          loanid (str): the id or encoded key of the loan account
          notes (str): notes to associate to the writeoff operation in Mambu
        """
        url = "https://{}/api/{}/{}:writeOff".format(self._tenant, "loans", loanid)
        data = {"notes": notes}
        return self.__request("POST", url, data=data)

    def mambu_change_state(self, entid, prefix, action, notes):
        """change state of mambu entity

        Args:
          entid (str): the id or encoded key of the entity owning the document
          prefix (str): entity's URL prefix
          action (str): specify the action state
          notes (str): notes to associate to the change of status
        """
        url = "https://{}/api/{}/{}:changeState".format(self._tenant, prefix, entid)
        data = {"action": action, "notes": notes}
        return self.__request("POST", url, data=data)

    def mambu_get_customfield(self, customfieldid):
        """Retrieves a Custom Field.

        Args:
          customfieldid (str): the id or encoded key of the custom field
        """
        url = "https://{}/api/customfields/{}".format(self._tenant, customfieldid)
        return self.__request("GET", url)

    def mambu_get_comments(
        self, owner_id, owner_type, offset=None, limit=None, paginationDetails="OFF"
    ):
        """Retrieves the comments of entity with owner_id.

        Args:
          owner_id (str): the id or encoded key of the owner
          owner_type (str): the type of the entity owning the comments
        """
        params = self.__validate_query_params(
            offset=offset,
            limit=limit,
            paginationDetails=paginationDetails,
        )
        if owner_type not in [
            "CLIENT",
            "GROUP",
            "LOAN_PRODUCT",
            "SAVINGS_PRODUCT",
            "CENTRE",
            "BRANCH",
            "USER",
            "LOAN_ACCOUNT",
            "DEPOSIT_ACCOUNT",
            "ID_DOCUMENT",
            "LINE_OF_CREDIT",
            "GL_JOURNAL_ENTRY",
        ]:
            raise MambuError("Owner {} not allowed".format(owner_type))

        params.update({"ownerType": owner_type})
        params.update({"ownerKey": owner_id})

        url = "https://{}/api/comments".format(self._tenant)

        return self.__request("GET", url, params=params)

    def mambu_comment(self, owner_id, owner_type, text):
        """Comments an entity with owner_id.

        Args:
          owner_id (str): the id or encoded key of the owner
          owner_type (str): the type of the entity owning the comments
          text (str): the text of the comment
        """
        if owner_type not in [
            "CLIENT",
            "GROUP",
            "LOAN_PRODUCT",
            "SAVINGS_PRODUCT",
            "CENTRE",
            "BRANCH",
            "USER",
            "LOAN_ACCOUNT",
            "DEPOSIT_ACCOUNT",
            "ID_DOCUMENT",
            "LINE_OF_CREDIT",
            "GL_JOURNAL_ENTRY",
        ]:
            raise MambuError("Owner {} not allowed".format(owner_type))
        data = {"ownerKey": owner_id, "ownerType": owner_type, "text": text}

        url = "https://{}/api/comments".format(self._tenant)

        return self.__request("POST", url, data=data)

    def mambu_make_disbursement(
        self, loan_id, notes, firstRepaymentDate, valueDate, allowed_fields, **kwargs
    ):
        """Make a disbursement transacton on a loan account.

        Args:
          loan_id (str): loan account id  to disburse
          firstRepaymentDate (str): first repayment date in ISO format
          notes (str): notes for the disbursement transaction
          valueDate (str): entrydate for disbursement transaction in ISO format
          allowed_fields (list): extra fields allowed for the transaction
          kwargs (dict): key-values of extra fields for the transaction
        """
        data = {
            "firstRepaymentDate": firstRepaymentDate,
            "notes": notes,
            "valueDate": valueDate,
        }
        for k, v in kwargs.items():
            if k in allowed_fields:
                data[k] = v

        url = "https://{}/api/loans/{}/disbursement-transactions".format(
            self._tenant, loan_id
        )

        return self.__request("POST", url, data=data)

    def mambu_make_repayment(
        self, loan_id, amount, notes, valueDate,
        allowed_fields, loantransaction_allowed_fields,
        **kwargs
    ):
        """Make a repayment transaction on a loan account.

        Args:
          loan_id (str): loan account id to make a repayment
          amount (float): the amount of the repayment
          notes (str): notes for the repayment transaction
          valueDate (str): date for the repayment transaction in ISO format
          allowed_fields (list): extra fields allowed for the transaction
          loantransaction_allowed_fields (list): extra fields allowed for the transaction details
          kwargs (dict): key-values of extra fields for the transaction
                         allows custom fields in the transaction ONLY when prefixed with '_' and being a dict
        """
        def _extract_valid_transactionDetails(transactionDetails_dict):
            transactionDetails = {}
            for k, v in transactionDetails_dict.items():
                if k in loantransaction_allowed_fields:
                    transactionDetails[k] = v
            return transactionDetails

        data = {"amount": amount, "notes": notes, "valueDate": valueDate}
        for k, v in kwargs.items():
            if k in allowed_fields:
                if k == "transactionDetails":
                    data[k] = _extract_valid_transactionDetails(v)
                else:
                    data[k] = v
            elif k[0] == "_" and isinstance(v, dict):
                data[k] = v

        url = "https://{}/api/loans/{}/repayment-transactions".format(
            self._tenant, loan_id
        )

        return self.__request("POST", url, data=data)

    def mambu_make_fee(
        self, loan_id, amount, installmentNumber, notes, valueDate, allowed_fields, **kwargs
    ):
        """Make a fee transaction on a loan account.

        Args:
          loan_id (str): loan account id to apply a fee
          amount (float): the amount of the fee
          installmentNumber (int): the installment number to apply the fee
          notes (str): notes for the fee transaction
          valueDate (str): entry date for the fee transaction
          allowed_fields (list): extra fields allowed for the transaction
          kwargs (dict): key-values of extra fields for the transaction
        """
        data = {
            "amount": amount,
            "installmentNumber": installmentNumber,
            "notes": notes,
            "valueDate": valueDate,
        }
        for k, v in kwargs.items():
            if k in allowed_fields:
                data[k] = v

        url = "https://{}/api/loans/{}/fee-transactions".format(self._tenant, loan_id)

        return self.__request("POST", url, data=data)

    def mambu_loantransaction_adjust(self, transactionid, notes):
        """Adjust a loan transaction

        Args:
          transactionid (str): the id or encoded key of the loan transaction
          notes (str): notes to associate to the transaction adjustment in Mambu
        """
        data = {
            "notes": notes,
        }

        url = "https://{}/api/loans/transactions/{}:adjust".format(self._tenant, transactionid)

        return self.__request("POST", url, data=data)



"""MambuLoan entity: a MambuEntity struct for credit Loans.

.. autosummary::
   :nosignatures:
   :toctree: _autosummary
"""

import copy
import datetime
import json

from dateutil.tz import tzlocal

from .entities import (
    MambuEntity,
    MambuEntityWritable,
    MambuEntityAttachable,
    MambuEntitySearchable,
    MambuEntityCommentable,
    MambuEntityOwnable,
    MambuInstallment,
)

from MambuPy.api.vos import (
    MambuDisbursementLoanTransactionInput,
    MambuFeeLoanTransactionInput,
    MambuRepaymentLoanTransactionInput,
    MambuLoanTransactionDetailsInput,
)
from MambuPy.mambuutil import MambuPyError
from .mambutransaction import MambuTransaction

class MambuLoan(
    MambuEntity,
    MambuEntityWritable,
    MambuEntityAttachable,
    MambuEntityCommentable,
    MambuEntityOwnable,
    MambuEntitySearchable,
):
    """MambuLoan entity"""

    _prefix = "loans"
    """prefix constant for connections to Mambu"""

    _filter_keys = [
        "branchId",
        "centreId",
        "accountState",
        "accountHolderType",
        "accountHolderId",
        "creditOfficerUsername",
    ]
    """allowed filters for get_all filtering"""

    _default_tzattrs = {
        "disbursementDetails": {
            "expectedDisbursementDate": None,
            "disbursementDate": None,
            "firstRepaymentDate": None,
        },
    }

    _sortBy_fields = [
        "creationDate",
        "lastModifiedDate",
        "id",
        "loanName",
    ]
    """allowed fields for get_all sorting"""

    _ownerType = "LOAN_ACCOUNT"
    """owner type of this entity"""

    _vos = [("disbursementDetails", "MambuDisbursementDetails")]
    """2-tuples of elements and Value Objects"""

    _entities = [
        ("assignedUserKey", "mambuuser.MambuUser", "assignedUser"),
        ("assignedBranchKey", "mambubranch.MambuBranch", "assignedBranch"),
        ("assignedCentreKey", "mambucentre.MambuCentre", "assignedCentre"),
        ("productTypeKey", "mambuproduct.MambuProduct", "productType"),
        ("originalAccountKey", "mambuloan.MambuLoan", "originalAccount"),
        ("accountHolderKey", "", "accountHolder"),
    ]
    """3-tuples of elements and Mambu Entities"""

    _accepted_actions = [
        "REQUEST_APPROVAL",
        "SET_INCOMPLETE",
        "APPROVE",
        "UNDO_APPROVE",
        "REJECT",
        "WITHDRAW",
        "CLOSE",
        "UNDO_REJECT",
        "UNDO_WITHDRAW",
        "UNDO_CLOSE",
    ]
    """accepted actions to change"""

    def __init__(self, **kwargs):
        self._entities = copy.deepcopy(MambuLoan._entities)
        super().__init__(**kwargs)
        self._attachments = {}

    def _delete_for_creation(self):
        """Deletes extra fields from Mambu unusable for entity
        creation."""
        try:
            del self._attrs["currency"]
            del self._attrs["accountState"]
            del self._attrs["scheduleSettings"]["hasCustomSchedule"]
            del self._attrs["interestSettings"]["accrueLateInterest"]
        except KeyError:
            pass

    def get_schedule(self, **kwargs):
        """Retrieves the installments schedule."""
        resp = self._connector.mambu_loanaccount_getSchedule(self.id)

        installments = json.loads(resp.decode())["installments"]

        self.schedule = []
        for installment in installments:
            installment_entity = MambuInstallment(**installment)
            installment_entity._tzattrs = copy.deepcopy(installment)
            installment_entity._convertDict2Attrs()
            self.schedule.append(installment_entity)

    def get_transactions(self, **kwargs):
        """Retrieves the transactions of the loan"""
        self.transactions = MambuTransaction.get_all(self.id)

    def set_state(self, action, notes, **kwargs):
        """Request to change status of a MambuLoan

        Args:
            action (str): specify the action state
            notes (str): notes to associate to the change of status

        Raises:
          `MambuPyError`: if action not in _accepted_actions
        """
        if action in self._accepted_actions:
            resp = self._connector.mambu_change_state(
                entid=self.id, prefix=self._prefix, action=action, notes=notes,
            )
            resp = json.loads(resp)
            self.accountState = resp["accountState"]
        else:
            raise MambuPyError(
                "field {} not in allowed _accepted_actions: {}".format(
                    action, self._accepted_actions
                )
            )

    def approve(self, notes, **kwargs):
        """Request to approve a loan account.

        Args:
          notes (str): notes to attach to the approval operation.
        """
        self.set_state("APPROVE", notes, **kwargs)

    def disburse(self, notes, firstRepaymentDate=None, disbursementDate=None, **kwargs):
        """Request to disburse a loan account.

        Args:
          notes (str): notes to attach to the disbursement transaction.
          firstRepaymentDate (:py:obj:`datetime`): first repayment date for the
                             loan account. If None, value is fetched from
                             disbursement details. If naive datetime, use TZ
                             info from tzattrs
          disbursementDate (:py:obj:`datetime`): disbursement date for the loan
                           account. If None, value is fetched from disbursement
                           details, expected disbursement date. If naive
                           datetime, use TZ info from tzattrs
          kwargs (dict): allowed extra params for the disbursement transaction
                 request. :py:obj:`MambuPy.api.vos.MambuDisbursementLoanTransactionInput._schema_fields`
                 has the allowed fields permitted for this operation
        """
        self._serializeFields()

        if firstRepaymentDate:
            if not firstRepaymentDate.tzinfo:
                timezone_firstRepaymentDate = datetime.timezone(
                    datetime.timedelta(
                        hours=int(
                            self.disbursementDetails._tzattrs["firstRepaymentDate"][-6:-3]
                        )
                    )
                )
                firstRepaymentDate = firstRepaymentDate.astimezone(
                    timezone_firstRepaymentDate
                )
            firstRepaymentDate = firstRepaymentDate.isoformat()
        else:
            firstRepaymentDate = self.disbursementDetails.firstRepaymentDate

        if disbursementDate:
            if not disbursementDate.tzinfo:
                timezone_disbursementDate = datetime.timezone(
                    datetime.timedelta(
                        hours=int(
                            self.disbursementDetails._tzattrs["expectedDisbursementDate"][
                                -6:-3
                            ]
                        )
                    )
                )
                disbursementDate = disbursementDate.astimezone(timezone_disbursementDate)
            disbursementDate = disbursementDate.isoformat()
        else:
            disbursementDate = self.disbursementDetails.expectedDisbursementDate

        self.disbursement_tr_resp = self._connector.mambu_make_disbursement(
            self.id,
            notes,
            firstRepaymentDate,
            disbursementDate,
            MambuDisbursementLoanTransactionInput._schema_fields,
            **kwargs
        )

        self.refresh()

    def reject(self, notes, **kwargs):
        """Request to reject a loan account.

        Args:
          notes (str): notes to attach to the reject operation.
        """
        self.set_state("REJECT", notes, **kwargs)

    def close(self, notes, **kwargs):
        """Request to close a loan account.

        Args:
          notes (str): notes to attach to the closing operation.
        """
        self.set_state("CLOSE", notes, **kwargs)

    def writeoff(self, notes, **kwargs):
        """Request to writeoff a loan account.

        Args:
          notes (str): notes to attach to the writting off operation.
        """
        self._connector.mambu_loanaccount_writeoff(self.id, notes)
        self.refresh()

    def repay(self, amount, notes, valueDate, **kwargs):
        """Request to repay a loan account.

        Args:
          amount (float): the amount of the repayment
          notes (str): notes to attach to the repayment transaction.
          valueDate (:py:obj:`datetime`): value date for the repayment
                    operation. If naive datetime, use TZ info from tzattrs.
          kwargs (dict): allowed extra params for the repayment transaction
                 request. :py:obj:`MambuPy.api.vos.MambuRepaymentLoanTransactionInput._schema_fields`
                 has the allowed fields permitted for this operation.
                 If transactionDetails is provided, it needs to have the
                 structure defined at
                 :py:obj:`MambuPy.api.vos.MambuLoanTransactionDetailsInput._schema_fields`.
        """
        valueDate = datetime.datetime.strptime(
            valueDate.strftime("%Y-%m-%d %H%M%S"), "%Y-%m-%d %H%M%S"
        )
        valueDate = valueDate.astimezone(tzlocal()).isoformat()

        self._connector.mambu_make_repayment(
            self.id,
            amount,
            notes,
            valueDate,
            MambuRepaymentLoanTransactionInput._schema_fields,
            MambuLoanTransactionDetailsInput._schema_fields,
            **kwargs
        )

        self.refresh()

    def adjust_transaction(self, transactionId, notes, **kwargs):
        """Request to adjust a loan transaction.

        Args:
          transactionId (str): the id of the transaction to adjust
          notes (str): notes to attach to the adjusting transaction.
        """
        if hasattr(self, "transactions"):
            if transactionId not in [t.id for t in self.transactions]:
                raise MambuPyError(
                    "transactionId '{}' not found in loan transactions".format(
                        transactionId
                    )
                )
        self._connector.mambu_loantransaction_adjust(transactionId, notes)
        self.refresh()

    def apply_fee(self, amount, installmentNumber, notes, valueDate, **kwargs):
        """Request to apply a fee to a loan account.

        Args:
          amount (float): the amount of the fee
          installmentNumber (int): the installment number to apply the fee
          notes (str): notes to attach to the fee transaction.
          valueDate (:py:obj:`datetime`): value date for the fee
                    transaction. If naive datetime, use TZ info from tzattrs.
          kwargs (dict): allowed extra params for the fee transaction
                 request. :py:obj:`MambuPy.api.vos.MambuFeeLoanTransactionInput._schema_fields`
                 has the allowed fields permitted for this operation
        """
        valueDate = datetime.datetime.strptime(
            valueDate.strftime("%Y-%m-%d %H%M%S"), "%Y-%m-%d %H%M%S"
        )
        valueDate = valueDate.astimezone(tzlocal()).isoformat()

        self._connector.mambu_make_fee(
            self.id,
            amount,
            installmentNumber,
            notes,
            valueDate,
            MambuFeeLoanTransactionInput._schema_fields,
            **kwargs
        )

        self.refresh()



# coding: utf-8
"""MambuPy REST module - DEPRECATED.

.. warning::

   DEPRECATED: This module uses Mambu API v1 which is no longer supported.
   This code is maintained only for backwards compatibility with legacy systems.
   It will be removed in a future version of MambuPy.

   Only mambuactivity is available in this legacy module.
"""
import warnings

warnings.warn(
    "MambuPy.rest module is deprecated and will be removed in a future version. "
    "Mambu API v1 is no longer supported by Mambu.",
    DeprecationWarning,
    stacklevel=2,
)

from .mambuactivity import (
    MambuActivity,
    MambuActivities,
    getactivitiesurl,
)

__all__ = [
    "MambuActivity",
    "MambuActivities",
    "getactivitiesurl",
]



"""Schema tables for Mambu Clients.

.. autosummary::
   :nosignatures:
   :toctree: _autosummary

.. todo:: this are just very basic schemas for clients. A lot of fields
          are missing.
"""

from sqlalchemy import (Column, DateTime, ForeignKey, Integer, Numeric, String,
                        Table)
from sqlalchemy.orm import relationship

from . import schema_orm as orm
from .schema_addresses import Address
from .schema_branches import Branch
from .schema_groups import Group, GroupRoleName

dbname = orm.dbname
session = orm.session
Base = orm.Base


class Client(Base):
    """Client table."""

    __tablename__ = "client"
    __table_args__ = {"schema": dbname, "keep_existing": True}

    # Columns
    encodedKey = Column(String, primary_key=True)
    id = Column(String, index=True, unique=True)
    firstName = Column(String)
    middleName = Column(String)
    lastName = Column(String)
    homePhone = Column(String)
    mobilePhone1 = Column(String)
    mobilePhone2 = Column(String)
    emailAddress = Column(String)
    gender = Column(String)
    state = Column(String)
    loanCycle = Column(Integer)
    birthDate = Column(DateTime)
    creationDate = Column(DateTime)
    activationDate = Column(DateTime)
    approvedDate = Column(DateTime)
    groups = relationship(
        Group, secondary=lambda: ClientsGroups, back_populates="clients"
    )

    # Relationships
    assignedBranchKey = Column(String, ForeignKey(Branch.encodedKey))
    branch = relationship("Branch", back_populates="clients")
    addresses = relationship(
        "Address",
        back_populates="client",
        foreign_keys="Address.parentKey",
        primaryjoin="Address.parentKey == Client.encodedKey",
    )
    customInformation = relationship(
        "CustomFieldValue",
        back_populates="client",
        foreign_keys="CustomFieldValue.parentKey",
        primaryjoin="CustomFieldValue.parentKey == Client.encodedKey",
    )
    loans = relationship(
        "LoanAccount",
        back_populates="holder_client",
        foreign_keys="LoanAccount.accountHolderKey",
        primaryjoin="LoanAccount.accountHolderKey == Client.encodedKey",
    )
    activities = relationship("Activity", back_populates="client")
    identificationDocuments = relationship(
        "IdentificationDocument", back_populates="client"
    )
    roles = relationship("GroupRole", back_populates="client")

    @property
    def name(self):
        return "{}{} {}".format(
            self.firstName.strip(),
            (" " + self.middleName.strip()) if self.middleName else "",
            self.lastName.strip(),
        )

    def __repr__(self):
        return "<Client(id={}, name={})>".format(self.id, self.name)


ClientsGroups = Table(
    "groupmember",
    Base.metadata,
    Column(
        "clientkey",
        Integer,
        ForeignKey(Client.encodedKey),
        nullable=False,
        primary_key=True,
        doc="Reference to client",
    ),
    Column(
        "groupkey",
        Integer,
        ForeignKey(Group.encodedKey),
        nullable=False,
        primary_key=True,
        doc="Reference to group",
    ),
    schema=dbname,
)


class GroupRole(Base):
    """GroupRole table.

    Association object for many-to-many relationship
    """

    __tablename__ = "grouprole"
    __table_args__ = {"schema": dbname, "keep_existing": True}

    encodedKey = Column(String, primary_key=True)

    clientKey = Column(
        String, ForeignKey(Client.encodedKey), nullable=False, doc="Reference to client"
    )
    groupKey = Column(
        String, ForeignKey(Group.encodedKey), nullable=False, doc="Reference to group"
    )
    groupRoleNameKey = Column(
        String,
        ForeignKey(GroupRoleName.encodedKey),
        nullable=False,
        doc="Reference to role name",
    )

    client = relationship(Client, back_populates="roles")
    group = relationship(Group, back_populates="roles")
    roleName = relationship(GroupRoleName, back_populates="roles")

    @property
    def rolename(self):
        return self.roleName.name

    def __repr__(self):
        return "<GroupRole(name={}, client={}, group={})>".format(
            self.roleName, self.client, self.group
        )


class IdentificationDocument(Base):
    """IdentificationDocument table.
    Related with client
    """

    __tablename__ = "identificationdocument"
    __table_args__ = {"schema": dbname, "keep_existing": True}

    encodedKey = Column(String, primary_key=True)
    documentId = Column(String)
    documentType = Column(String)
    indexInList = Column(Integer)
    issuingAuthority = Column(String)
    validUntil = Column(DateTime)
    identificationDocumentTemplateKey = Column(String)

    # Relationships
    clientKey = Column(String, ForeignKey(Client.encodedKey))
    client = relationship(Client, back_populates="identificationDocuments")

    def __repr__(self):
        return "<IdentificationDocument(documentId={})>".format(self.documentId)



"""Schema tables for Mambu Users.

.. autosummary::
   :nosignatures:
   :toctree: _autosummary

.. todo:: this are just very basic schemas for users.
"""

from sqlalchemy import Column, DateTime, ForeignKey, Integer, Numeric, String
from sqlalchemy.orm import relationship

from . import schema_orm as orm
from .schema_branches import Branch

dbname = orm.dbname
session = orm.session
Base = orm.Base


class Role(Base):
    """Role table.
    Related with user
    """

    __tablename__ = "role"
    __table_args__ = {"schema": dbname, "keep_existing": True}

    # Columns
    encodedKey = Column(String, primary_key=True)
    name = Column(String)
    users = relationship("User", back_populates="role")

    def __repr__(self):
        return "<Role(name={})>".format(self.name)


class User(Base):
    """User table."""

    __tablename__ = "user"
    __table_args__ = {"schema": dbname, "keep_existing": True}

    # Columns
    encodedKey = Column(String, primary_key=True)
    id = Column(String, index=True, unique=True)
    firstName = Column(String)
    lastName = Column(String)
    userName = Column(String)
    homePhone = Column(String)
    mobilePhone1 = Column(String)
    userState = Column(String)
    email = Column(String)
    creationDate = Column(DateTime)
    isCreditOfficer = Column(Integer)

    # Relationships
    assignedBranchKey = Column(String, ForeignKey(Branch.encodedKey))
    branch = relationship(Branch, back_populates="users")

    role_encodedKey_oid = Column(String, ForeignKey(Role.encodedKey))
    role = relationship(Role, back_populates="users")

    customInformation = relationship(
        "CustomFieldValue",
        back_populates="user",
        foreign_keys="CustomFieldValue.parentKey",
        primaryjoin="CustomFieldValue.parentKey == User.encodedKey",
    )
    activities = relationship(
        "Activity",
        primaryjoin="User.encodedKey==Activity.userKey",
        back_populates="user",
    )
    assignedActivities = relationship(
        "Activity",
        primaryjoin="User.encodedKey==Activity.assignedUserKey",
        back_populates="assignedUser",
    )
    createdTasks = relationship(
        "Task",
        primaryjoin="User.encodedKey==Task.createdByUserKey",
        back_populates="createdByUser",
    )
    assignedTasks = relationship(
        "Task",
        primaryjoin="User.encodedKey==Task.assignedUserKey",
        back_populates="assignedUser",
    )
    groups = relationship("Group", back_populates="user")
    loans = relationship("LoanAccount", back_populates="user")

    @property
    def name(self):
        return "{} {}".format(self.firstName, self.lastName)

    def __repr__(self):
        return "<User(id={}, name={})>".format(self.id, self.name)



"""MambuRole entity: a MambuEntity struct for User Roles.

.. autosummary::
   :nosignatures:
   :toctree: _autosummary
"""

from .entities import MambuEntity, MambuEntityWritable


class MambuRole(MambuEntity, MambuEntityWritable):
    """MambuRole entity"""

    _prefix = "userroles"
    """prefix constant for connections to Mambu"""

    def __repr__(self):
        """Prints class and name of role"""
        try:
            return self.__class__.__name__ + " - name: %s" % self._attrs["name"]
        except Exception:
            return super().__repr__()



"""Interfaces for Mambu Objects.

.. autosummary::
   :nosignatures:
   :toctree: _autosummary
"""

from abc import ABC, abstractmethod


class MambuWritable(ABC):
    """An entity which allows basic write operations"""

    @abstractmethod
    def update(self):  # pragma: no cover
        """updates a mambu entity

        Uses the current values of the _attrs to send to Mambu.
        Pre-requires that CustomFields are updated previously.
        Post-requires that CustomFields are extracted again.
        """
        raise NotImplementedError

    @abstractmethod
    def create(self):  # pragma: no cover
        """creates a mambu entity

        Uses the current values of the _attrs to send to Mambu.
        Pre-requires that CustomFields are updated previously.
        Post-requires that CustomFields are extracted again.
        """
        raise NotImplementedError

    @abstractmethod
    def patch(self, fields=None, autodetect_remove=False):  # pragma: no cover
        """patches a mambu entity

        Allows patching of parts of the entity up to Mambu.

        fields is a list of the values in the _attrs that will be sent to Mambu

        autodetect automatically searches for deleted fields and patches a
        remove in Mambu.

        Pre-requires that CustomFields are updated previously.
        Post-requires that CustomFields are extracted again.

        Args:
          fields (list of str): list of ids of fields to explicitly patch
          autodetect_remove (bool): False: if deleted fields, don't remove them
                                    True: if delete field, remove them

        Autodetect operation, for any field (every field if fields param is
        None):
          * ADD: in attrs, but not in resp
          * REPLACE: in attrs, and in resp
          * REMOVE: not in attrs, but in resp
          * MOVE: not yet implemented (how to autodetect? request needs 'from' element)

        Raises:
          `MambuPyError`: if field not in attrrs, and not in resp
        """
        raise NotImplementedError


class MambuAttachable(ABC):
    """An entity which can attach documents"""

    @abstractmethod
    def attach_document(self, filename, title="", notes=""):  # pragma: no cover
        """uploads an attachment to this entity

        Args:
          filename (str): path and filename of file to upload as attachment
          title (str): name to assign to the attached file in Mambu
          notes (str): notes to associate to the attached file in Mambu

        Returns:
          Mambu's response with metadata of the attached document
        """
        raise NotImplementedError


class MambuSearchable(ABC):
    """An entity which allows searching endpoint"""

    @classmethod
    @abstractmethod
    def search(
        cls,
        filterCriteria=None,
        sortingCriteria=None,
        offset=None,
        limit=None,
        paginationDetails="OFF",
        detailsLevel="BASIC",
    ):  # pragma: no cover
        """search, several entities, filtering criteria allowed

        Args:
          filterCriteria (list of dicts): fields according to
                              LoanAccountFilterCriteria schema
          sortingCriteria (dict): fields according to
                            LoanAccountSortingCriteria
          offset (int): pagination, index to start searching
          limit (int): pagination, number of elements to retrieve
          paginationDetails (str ON/OFF): ask for details on pagination
          detailsLevel (str BASIC/FULL): ask for extra details or not

        Returns:
          list of instances of an entity with data from Mambu
        """
        raise NotImplementedError


class MambuCommentable(ABC):
    """An entity which allows comments endpoints"""

    def get_comments(self):  # pragma: no cover
        """Gets comments for this entity

        _comments list is cleaned and set with retrieved comments

        Args:
          offset (int): pagination, index to start searching
          limit (int): pagination, number of elements to retrieve
          paginationDetails (str ON/OFF): ask for details on pagination

        Returns:
          Mambu's response with retrieved comments
        """
        raise NotImplementedError

    def comment(self, comment):  # pragma: no cover
        raise NotImplementedError


class MambuOwner(ABC):
    """"""


class MambuOwnable(ABC):
    """An entity which allows to be 'owned' by another.

    An owned entity has an 'accountHolderKey' and 'accountHolderType'
    fields.

    Because of that, you may call get_accountHolder on the owned
    entity to instantiate the MambuEntity who owns it.
    """

    def _assignEntObjs(
        self, entities, detailsLevel, get_entities, debug
    ):  # pragma: no cover
        raise NotImplementedError


class MambuHolder(ABC):
    """"""



"""Schema tables for Mambu Groups.

.. autosummary::
   :nosignatures:
   :toctree: _autosummary

.. todo:: this are just very basic schemas for groups. A some fields
          are missing.
"""

from . import schema_orm as orm
from .schema_branches import Branch
from .schema_centres import Centre
from .schema_users import User

# technical debt, cannot import name schema_clients
try:
    from . import schema_clients
except Exception:
    import schema_clients

from sqlalchemy import (Column, DateTime, ForeignKey, Integer, Numeric, String,
                        Text)
from sqlalchemy.orm import relationship

dbname = orm.dbname
session = orm.session
Base = orm.Base


class Group(Base):
    """Group table."""

    __tablename__ = "group"
    __table_args__ = {"schema": dbname, "keep_existing": True}

    # Columns
    encodedKey = Column(String, primary_key=True)
    id = Column(String, index=True, unique=True)
    groupName = Column(String)
    homePhone = Column(String)
    mobilePhone1 = Column(String)
    loanCycle = Column(Integer)
    creationDate = Column(DateTime)
    notes = Column(Text)

    # Relationships
    assignedCentreKey = Column(String, ForeignKey(Centre.encodedKey))
    assignedUserKey = Column(String, ForeignKey(User.encodedKey))
    assignedBranchKey = Column(String, ForeignKey(Branch.encodedKey))
    user = relationship(User, back_populates="groups")
    centre = relationship(Centre, back_populates="groups")
    branch = relationship(Branch, back_populates="groups")
    addresses = relationship(
        "Address",
        back_populates="group",
        foreign_keys="Address.parentKey",
        primaryjoin="Address.parentKey == Group.encodedKey",
    )
    customInformation = relationship(
        "CustomFieldValue",
        back_populates="group",
        foreign_keys="CustomFieldValue.parentKey",
        primaryjoin="CustomFieldValue.parentKey == Group.encodedKey",
    )
    loans = relationship(
        "LoanAccount",
        back_populates="holder_group",
        foreign_keys="LoanAccount.accountHolderKey",
        primaryjoin="LoanAccount.accountHolderKey == Group.encodedKey",
    )
    activities = relationship("Activity", back_populates="group")
    tasks = relationship(
        "Task",
        back_populates="link_group",
        foreign_keys="Task.taskLinkKey",
        primaryjoin="Task.taskLinkKey == Group.encodedKey",
        order_by="Task.dueDate",
    )
    clients = relationship(
        "Client",
        secondary=lambda: schema_clients.ClientsGroups,
        back_populates="groups",
    )
    roles = relationship("GroupRole", back_populates="group")

    def __repr__(self):
        return "<Group(id={}, groupName={})>".format(self.id, self.groupName)


class GroupRoleName(Base):
    """GroupRoleName table."""

    __tablename__ = "grouprolename"
    __table_args__ = {"schema": dbname, "keep_existing": True}

    # Columns
    encodedKey = Column(String, primary_key=True)
    name = Column(String)

    roles = relationship("schema_clients.GroupRole", back_populates="roleName")

    def __repr__(self):
        return "<GroupRoleName(name={})>".format(self.name)



"""MambuUser entity: a MambuEntity struct for Users.

.. autosummary::
   :nosignatures:
   :toctree: _autosummary
"""

from .entities import (
    MambuEntity,
    MambuEntityWritable,
    MambuEntityAttachable,
    MambuEntityCommentable,
    MambuPyError,
)
from .interfaces import MambuOwner


class MambuUser(
    MambuEntity,
    MambuEntityWritable,
    MambuEntityAttachable,
    MambuEntityCommentable,
    MambuOwner,
):
    """MambuUser entity"""

    _prefix = "users"
    """prefix constant for connections to Mambu"""

    _filter_keys = [
        "branchId",
    ]
    """allowed filters for get_all filtering"""

    _sortBy_fields = []
    """allowed fields for get_all sorting"""

    _ownerType = "USER"
    """owner type of this entity"""

    _entities = [
        ("assignedBranchKey", "mambubranch.MambuBranch", "assignedBranch"),
        ("role", "mamburole.MambuRole", "role"),
    ]
    """3-tuples of elements and Mambu Entities"""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._attachments = {}

    def __repr__(self):
        """Prints class and name of role"""
        try:
            return self.__class__.__name__ + " - username: %s" % self._attrs["username"]
        except Exception:
            return super().__repr__()

    def __getattribute__(self, name):
        """Object-like get attribute for MambuUsers."""
        if name == "get_role":
            return lambda **kwargs: self.getEntities(entities=["role"], **kwargs)[0]
        else:
            return super().__getattribute__(name)

    def _updateVOs(self):
        """Updates role, and the loops _vos list to update corresponding data"""
        self._attrs["role"] = {
            "encodedKey": self.role["encodedKey"],
            "id": self.role["id"],
        }
        super()._updateVOs()

    @classmethod
    def get_all(
        cls,
        filters=None,
        offset=None,
        limit=None,
        paginationDetails="OFF",
        detailsLevel="BASIC",
        sortBy=None,
        branchIdType=None,
        **kwargs,
    ):
        """get_all, several entities, filtering allowed

        Args:
          filters (dict): key-value filters, dependes on each entity
                          (keys must be one of the _filter_keys)
          offset (int): pagination, index to start searching
          limit (int): pagination, number of elements to retrieve
          paginationDetails (str ON/OFF): ask for details on pagination
          detailsLevel (str BASIC/FULL): ask for extra details or not
          sortBy (str): ``field1:ASC,field2:DESC``, sorting criteria for
                        results (fields must be one of the _sortBy_fields)
          branchIdType (str ASSIGNED/MANAGE): search assigned users or also
                       those who can manage a branch

        Returns:
          list of instances of an entity with data from Mambu
        """
        if branchIdType and (not filters or "branchId" not in filters):
            raise MambuPyError("branchIdType not allowed if branchId not provided")
        if branchIdType and branchIdType not in ["ASSIGNED", "MANAGE"]:
            raise MambuPyError("Invalid branchIdType: {}".format(branchIdType))

        params = {}
        if branchIdType:
            params["branchIdType"] = branchIdType
        params.update(kwargs)

        return super().get_all(
            filters, offset, limit, paginationDetails, detailsLevel, sortBy, **params
        )
